
import datetime as _datetime
import json as _json
import os as _os

__all__ = ["PAR", "BucketReader", "BucketWriter", "ObjectReader",
           "ObjectWriter", "ComputeRunner"]


class PAR:
    """This class holds the result of a pre-authenticated request
       (a PAR - also called a pre-signed request). This holds a
       pre-authenticated URL to access either;

       (1) A individual object in an object store (read or write)
       (2) An entire bucket in the object store (write only)
       (3) A calculation to be performed on the compute service (start or stop)

       The PAR is created encrypted, so can only be used by the
       person or service that has access to the decryption key
    """
    def __init__(self, url=None, key=None, encrypt_key=None,
                 created_datetime=None,
                 expires_datetime=None,
                 is_readable=False,
                 is_writeable=False,
                 is_executable=False,
                 par_id=None, par_name=None,
                 driver=None):
        """Construct a PAR result by passing in the URL at which the
           object can be accessed, the UTC datetime when this expires,
           whether this is readable, writeable or executable, and
           the encryption key to use to encrypt the PAR.

           If this is an object store PAR, then optionally you can
           pass in the key for the object in the object store that
           this provides access to. If this is not supplied, then an
           entire bucket is accessed). If 'is_readable', then read-access
           has been granted, while if 'is_writeable' then write
           access has been granted.

           If 'is_executable' then this is a calculation PAR that triggers
           a calculation.

           Otherwise no access is possible.

           This also records the type of object store behind this PAR
           in the free-form string 'driver'. You can optionally supply
           the ID of the PAR by passing in 'par_id', the user-supplied name,
           of the PAR by passing in 'par_name', and the time it
           was created using 'created_datetime' (in the same format
           as 'expires_datetime' - should be a UTC datetime with UTC tzinfo)
        """
        if url is None:
            is_readable = True
            self._uid = None
        else:
            from Acquire.Crypto import PublicKey as _PublicKey
            if not isinstance(encrypt_key, _PublicKey):
                raise TypeError(
                    "You must supply a valid PublicKey to encrypt a PAR")

            url = encrypt_key.encrypt(url)

            from Acquire.ObjectStore import create_uuid as _create_uuid
            self._uid = _create_uuid()

        self._url = url
        self._key = key
        self._created_datetime = created_datetime
        self._expires_datetime = expires_datetime
        self._driver = driver
        self._par_id = par_id
        self._par_name = par_name

        if is_readable:
            self._is_readable = True
        else:
            self._is_readable = False

        if is_writeable:
            self._is_writeable = True
        else:
            self._is_writeable = False

        if is_executable:
            self._is_executable = True
        else:
            self._is_executable = False

        if self._is_executable:
            self._is_readable = False
            self._is_writeable = False
        elif not (self._is_readable or self._is_writeable):
            from Acquire.Client import PARPermissionsError
            raise PARPermissionsError(
                "You cannot create a PAR that has no read or write "
                "or execute permissions!")
        else:
            self._is_executable = False

    def __str__(self):
        if self.seconds_remaining() < 1:
            return "PAR( expired )"
        elif self._is_executable:
            return "PAR( calculation, seconds_remaining=%s )" % \
                (self.seconds_remaining(buffer=0))
        if self._key is None:
            return "PAR( bucket=True, seconds_remaining=%s )" % \
                (self.seconds_remaining(buffer=0))
        else:
            return "PAR( key=%s, seconds_remaining=%s )" % \
                (self.key(), self.seconds_remaining(buffer=0))

    def is_null(self):
        """Return whether or not this is null"""
        return self._uid is None

    def url(self, encrypt_key):
        """Return the URL at which the bucket/object can be accessed. This
           will raise a PARTimeoutError if the url has less than 30 seconds
           of validity left. Note that you must pass in the key used to
           decrypt the PAR"""
        if self.seconds_remaining(buffer=30) <= 0:
            from Acquire.Client import PARTimeoutError
            raise PARTimeoutError(
                "The URL behind this PAR has expired and is no longer valid")

        return encrypt_key.decrypt(self._url)

    def uid(self):
        """Return the UID of this PAR"""
        return self._uid

    def par_id(self):
        """Return the ID of the PAR, if this was supplied by the underlying
           driver. This could be useful for PAR management by the server
        """
        return self._par_id

    def par_name(self):
        """Return the user-supplied name of the PAR, if this was supplied
           by the user and supported by the underlying driver. This could
           be useful for PAR management by the server
        """
        return self._par_name

    def fingerprint(self):
        """Return a fingerprint for this PAR that can be used
           in authorisations
        """
        return self._uid

    def is_readable(self):
        """Return whether or not this PAR gives read access"""
        return self._is_readable

    def is_writeable(self):
        """Return whether or not this PAR gives write access"""
        return self._is_writeable

    def is_executable(self):
        """Return whether or not this is an executable job"""
        return self._is_executable

    def key(self):
        """Return the key for the object this accesses - this is None
           if the PAR grants access to the entire bucket"""
        return self._key

    def is_bucket(self):
        """Return whether or not this PAR is for an entire bucket"""
        return (self._key is None) and not (self.is_executable())

    def is_calculation(self):
        """Return whether or not this PAR is for a calculation"""
        return self._is_executable

    def is_object(self):
        """Return whether or not this PAR is for a single object"""
        return self._key is not None

    def driver(self):
        """Return the underlying object store driver used for this PAR"""
        return self._driver

    def seconds_remaining(self, buffer=30):
        """Return the number of seconds remaining before this PAR expires.
           This will return 0 if the PAR has already expired. To be safe,
           you should renew PARs if the number of seconds remaining is less
           than 60. This will subtract 'buffer' seconds from the actual
           validity to provide a buffer against race conditions (function
           says this is valid when it is not)
        """
        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now
        now = _get_datetime_now()

        buffer = float(buffer)

        if buffer < 0:
            buffer = 0

        delta = (self._expires_datetime - now).total_seconds() - buffer

        if delta < 0:
            return 0
        else:
            return delta

    def read(self, encrypt_key):
        """Return an object that can be used to read data from this PAR"""
        if not self.is_readable():
            from Acquire.Client import PARPermissionsError
            raise PARPermissionsError(
                "You do not have permission to read from this PAR: %s" % self)

        if self.is_bucket():
            return BucketReader(self, encrypt_key)
        else:
            return ObjectReader(self, encrypt_key)

    def write(self, encrypt_key):
        """Return an object that can be used to write data to this PAR"""
        if not self.is_writeable():
            from Acquire.Client import PARPermissionsError
            raise PARPermissionsError(
                "You do not have permission to write to this PAR: %s" % self)

        if self.is_bucket():
            return BucketWriter(self, encrypt_key)
        else:
            return ObjectWriter(self, encrypt_key)

    def execute(self, encrypt_key):
        """Return an object that can be used to control execution of
           this PAR
        """
        if not self.is_executable():
            from Acquire.Client import PARPermissionsError
            raise PARPermissionsError(
                "You do not have permission to execute this PAR: %s" % self)

        return ComputeRunner(self, encrypt_key)

    def close(self):
        """Close this PAR - this closes and deletes the PAR"""
        pass

    def to_data(self):
        """Return a json-serialisable dictionary that contains all data
           for this object
        """
        data = {}

        if self._url is None:
            return data

        from Acquire.ObjectStore import datetime_to_string \
            as _datetime_to_string
        from Acquire.ObjectStore import bytes_to_string \
            as _bytes_to_string

        data["url"] = _bytes_to_string(self._url)
        data["uid"] = self._uid
        data["key"] = self._key
        data["created_datetime"] = _datetime_to_string(self._created_datetime)
        data["expires_datetime"] = _datetime_to_string(self._expires_datetime)
        data["driver"] = self._driver
        data["par_id"] = self._par_id
        data["par_name"] = self._par_name
        data["is_readable"] = self._is_readable
        data["is_writeable"] = self._is_writeable
        data["is_executable"] = self._is_executable

        return data

    @staticmethod
    def from_data(data):
        """Return a PAR constructed from the passed json-deserliased
           dictionary
        """
        if data is None or len(data) == 0:
            return PAR()

        from Acquire.ObjectStore import string_to_datetime \
            as _string_to_datetime
        from Acquire.ObjectStore import string_to_bytes \
            as _string_to_bytes

        par = PAR()

        par._url = _string_to_bytes(data["url"])
        par._key = data["key"]
        par._uid = data["uid"]

        if par._key is not None:
            par._key = str(par._key)

        par._created_datetime = _string_to_datetime(data["created_datetime"])
        par._expires_datetime = _string_to_datetime(data["expires_datetime"])
        par._driver = data["driver"]
        par._par_id = data["par_id"]
        par._par_name = data["par_name"]
        par._is_readable = data["is_readable"]
        par._is_writeable = data["is_writeable"]
        par._is_executable = data["is_executable"]

        return par


def _url_to_filepath(url):
    """Internal function used to strip the "file://" from the beginning
       of a file url
    """
    return url[7:]


def _read_local(url):
    """Internal function used to read data from the local testing object
       store
    """
    with open("%s._data" % _url_to_filepath(url), "rb") as FILE:
        return FILE.read()


def _read_remote(url):
    """Internal function used to read data from a remote URL"""
    status_code = None
    response = None

    try:
        from Acquire.Stubs import requests as _requests
        response = _requests.get(url)
        status_code = response.status_code
    except Exception as e:
        from Acquire.Client import PARReadError
        raise PARReadError(
            "Cannot read the remote PAR URL '%s' because of a possible "
            "nework issue: %s" % (url, str(e)))

    output = response.content

    if status_code != 200:
        from Acquire.Client import PARReadError
        raise PARReadError(
            "Failed to read data from the PAR URL. HTTP status code = %s, "
            "returned output: %s" % (status_code, output))

    return output


def _list_local(url):
    """Internal function to list all of the objects keys below 'url'"""
    local_dir = _url_to_filepath(url)

    keys = []

    for dirpath, _, filenames in _os.walk(local_dir):
        local_path = dirpath[len(local_dir):]
        has_local_path = (len(local_path) > 0)

        for filename in filenames:
            if filename.endswith("._data"):
                filename = filename[0:-6]

                if has_local_path:
                    keys.append("%s/%s" % (local_path, filename))
                else:
                    keys.append(filename)

    return keys


def _list_remote(url):
    """Internal function to list all of the objects keys below 'url'"""
    return []


def _write_local(url, data):
    """Internal function used to write data to a local file"""
    filename = "%s._data" % _url_to_filepath(url)

    try:
        with open(filename, 'wb') as FILE:
            FILE.write(data)
            FILE.flush()
    except:
        dir = "/".join(filename.split("/")[0:-1])
        _os.makedirs(dir, exist_ok=True)
        with open(filename, 'wb') as FILE:
            FILE.write(data)
            FILE.flush()


def _write_remote(url, data):
    """Internal function used to write data to the passed remote URL"""
    try:
        from Acquire.Stubs import requests as _requests
        response = _requests.put(url, data=data)
        status_code = response.status_code
    except Exception as e:
        from Acquire.Client import PARWriteError
        raise PARWriteError(
            "Cannot write data to the remote PAR URL '%s' because of a "
            "possible nework issue: %s" % (url, str(e)))

    if status_code != 200:
        from Acquire.Client import PARWriteError
        raise PARWriteError(
            "Cannot write data to the remote PAR URL '%s' because of a "
            "possible nework issue: %s" % (url, str(response.content)))


def _join_bucket_and_prefix(url, prefix):
    """Join together the passed url and prefix, returning the
       url directory and the remainig part which is the start
       of the file name
    """
    if prefix is None:
        return url

    parts = prefix.split("/")

    return ("%s/%s" % (url, "/".join(parts[0:-2])), parts[-1])


class BucketReader:
    """This class provides functions to enable reading data from a
       bucket via a PAR
    """
    def __init__(self, par=None, encrypt_key=None):
        if par:
            if not isinstance(par, PAR):
                raise TypeError(
                    "You can only create a BucketReader from a PAR")
            elif not par.is_bucket():
                raise ValueError(
                    "You can only create a BucketReader from a PAR that "
                    "represents an entire bucket: %s" % par)
            elif not par.is_readable():
                from Acquire.Client import PARPermissionsError
                raise PARPermissionsError(
                    "You cannot create a BucketReader from a PAR without "
                    "read permissions: %s" % par)

            self._par = par
            self._url = par.url(encrypt_key)
        else:
            self._par = None

    def get_object(self, key):
        """Return the binary data contained in the key 'key' in the
           passed bucket"""
        if self._par is None:
            from Acquire.Client import PARError
            raise PARError("You cannot read data from an empty PAR")

        while key.startswith("/"):
            key = key[1:]

        url = self._url

        if url.endswith("/"):
            url = "%s%s" % (url, key)
        else:
            url = "%s/%s" % (url, key)

        if url.startswith("file://"):
            return _read_local(url)
        else:
            return _read_remote(url)

    def get_object_as_file(self, key, filename):
        """Get the object contained in the key 'key' in the passed 'bucket'
           and writing this to the file called 'filename'"""
        objdata = self.get_object(key)

        with open(filename, "wb") as FILE:
            FILE.write(objdata)

    def get_string_object(self, key):
        """Return the string in 'bucket' associated with 'key'"""
        data = self.get_object(key)

        try:
            return data.decode("utf-8")
        except Exception as e:
            raise TypeError(
                "The object behind this PAR cannot be converted to a string. "
                "Error is: %s" % str(e))

    def get_object_from_json(self, key):
        """Return an object constructed from json stored at 'key' in
           the passed bucket. This raises an exception if there is no
           data or the PAR has expired
        """
        data = self.get_string_object(key)
        return _json.loads(data)

    def get_all_object_names(self, prefix=None):
        """Returns the names of all objects in the passed bucket"""
        (url, part) = _join_bucket_and_prefix(self._url, prefix)

        if url.startswith("file://"):
            objnames = _list_local(url)
        else:
            objnames = _list_remote(url)

        # scan the object names returned and discard the ones that don't
        # match the prefix
        matches = []

        if len(part) > 0:
            for objname in objnames:
                if objname.startswith(part):
                    objname = objname[len(part):]

                    while objname.startswith("/"):
                        objname = objname[1:]

                    matches.append(objname)
        else:
            matches = objnames

        return matches

    def get_all_objects(self, prefix=None):
        """Return all of the objects in the passed bucket"""
        names = self.get_all_object_names(prefix)

        objects = {}

        if prefix:
            for name in names:
                objects[name] = self.get_object(
                                    "%s/%s" % (prefix, name))
        else:
            for name in names:
                objects[name] = self.get_object(name)

        return objects

    def get_all_strings(self, prefix=None):
        """Return all of the strings in the passed bucket"""
        objects = self.get_all_objects(prefix)

        names = list(objects.keys())

        for name in names:
            try:
                s = objects[name].decode("utf-8")
                objects[name] = s
            except:
                del objects[name]

        return objects


class BucketWriter:
    """This class provides functions to enable writing data to a
       bucket via a PAR
    """
    def __init__(self, par=None, encrypt_key=None):
        if par:
            if not isinstance(par, PAR):
                raise TypeError(
                    "You can only create a BucketReader from a PAR")
            elif not par.is_bucket():
                raise ValueError(
                    "You can only create a BucketReader from a PAR that "
                    "represents an entire bucket: %s" % par)
            elif not par.is_writeable():
                from Acquire.Client import PARPermissionsError
                raise PARPermissionsError(
                    "You cannot create a BucketWriter from a PAR without "
                    "write permissions: %s" % par)

            self._par = par
            self._url = par.url(encrypt_key)
        else:
            self._par = None

    def set_object(self, key, data):
        """Set the value of 'key' in 'bucket' to binary 'data'"""
        if self._par is None:
            from Acquire.Client import PARError
            raise PARError("You cannot write data to an empty PAR")

        while key.startswith("/"):
            key = key[1:]

        url = self._url

        if url.endswith("/"):
            url = "%s%s" % (url, key)
        else:
            url = "%s/%s" % (url, key)

        if url.startswith("file://"):
            return _write_local(url, data)
        else:
            return _write_remote(url, data)

    def set_object_from_file(self, key, filename):
        """Set the value of 'key' in 'bucket' to equal the contents
           of the file located by 'filename'"""
        with open(filename, "rb") as FILE:
            data = FILE.read()
            self.set_object(key, data)

    def set_string_object(self, key, string_data):
        """Set the value of 'key' in 'bucket' to the string 'string_data'"""
        self.set_object(key, string_data.encode("utf-8"))

    def set_object_from_json(self, key, data):
        """Set the value of 'key' in 'bucket' to equal to contents
           of 'data', which has been encoded to json"""
        self.set_string_object(key, _json.dumps(data))


class ObjectReader:
    """This class provides functions for reading an object via a PAR"""
    def __init__(self, par=None, encrypt_key=None):
        if par:
            if not isinstance(par, PAR):
                raise TypeError(
                    "You can only create an ObjectReader from a PAR")
            elif par.is_bucket():
                raise ValueError(
                    "You can only create an ObjectReader from a PAR that "
                    "represents an object: %s" % par)
            elif not par.is_readable():
                from Acquire.Client import PARPermissionsError
                raise PARPermissionsError(
                    "You cannot create an ObjectReader from a PAR without "
                    "read permissions: %s" % par)

            self._par = par
            self._url = par.url(encrypt_key)
        else:
            self._par = None

    def get_object(self):
        """Return the binary data contained in this object"""
        if self._par is None:
            from Acquire.Client import PARError
            raise PARError("You cannot read data from an empty PAR")

        url = self._url

        if url.startswith("file://"):
            return _read_local(url)
        else:
            return _read_remote(url)

    def get_object_as_file(self, filename):
        """Get the object contained in this PAR and write this to
           the file called 'filename'"""
        objdata = self.get_object()

        with open(filename, "wb") as FILE:
            FILE.write(objdata)

    def get_string_object(self):
        """Return the object behind this PAR as a string (raises exception
           if it is not a string)'"""
        data = self.get_object()

        try:
            return data.decode("utf-8")
        except Exception as e:
            raise TypeError(
                "The object behind this PAR cannot be converted to a string. "
                "Error is: %s" % str(e))

    def get_object_from_json(self):
        """Return an object constructed from json stored at behind
           this PAR. This raises an exception if there is no data
           or the PAR has expired
        """
        data = self.get_string_object()
        return _json.loads(data)


class ObjectWriter(ObjectReader):
    """This is an extension of ObjectReader that also allows writing to
       the object via the PAR
    """
    def __init__(self, par=None, encrypt_key=None):
        if par:
            if not isinstance(par, PAR):
                raise TypeError(
                    "You can only create an ObjectReader from a PAR")
            elif par.is_bucket():
                raise ValueError(
                    "You can only create an ObjectReader from a PAR that "
                    "represents an object: %s" % par)
            elif not par.is_writeable():
                from Acquire.Client import PARPermissionsError
                raise PARPermissionsError(
                    "You cannot create an ObjectWriter from a PAR without "
                    "write permissions: %s" % par)

            self._par = par
            self._url = par.url(encrypt_key)
        else:
            self._par = None

    def set_object(self, data):
        """Set the value of the object behind this PAR to the binary 'data'"""
        if self._par is None:
            from Acquire.Client import PARError
            raise PARError("You cannot write data to an empty PAR")

        url = self._url

        if url.startswith("file://"):
            return _write_local(url, data)
        else:
            return _write_remote(url, data)

    def set_object_from_file(self, filename):
        """Set the value of the object behind this PAR to equal the contents
           of the file located by 'filename'"""
        with open(filename, "rb") as FILE:
            data = FILE.read()
            self.set_object(data)

    def set_string_object(self, string_data):
        """Set the value of the object behind this PAR to the
           string 'string_data'
        """
        self.set_object(string_data.encode("utf-8"))

    def set_object_from_json(self, data):
        """Set the value of the object behind this PAR to equal to contents
           of 'data', which has been encoded to json"""
        self.set_string_object(_json.dumps(data))


class ComputeRunner:
    """This class provides functions for executing a calculation
       pre-authorised by the passed PAR
    """
    def __init__(self, par=None, encrypt_key=None):
        if par:
            if not isinstance(par, PAR):
                raise TypeError(
                    "You can only create a ComputeRunner from a PAR")
            elif not par.is_executable():
                raise ValueError(
                    "You can only create a ComputeRunner from a PAR that "
                    "represents an executable calculation: %s" % par)

            self._par = par
            self._url = par.url(encrypt_key)
        else:
            self._par = None




__all__ = ["get_users_public_certs"]


def get_users_public_certs(identity_url, username, session_uid):
    """Call the identity_url to obtain the public keys
       and certificates of the user with 'username' logged
       in using the specified session_uid
    """
    from Acquire.Service import get_trusted_service as _get_trusted_service

    service = _get_trusted_service(identity_url)

    function = "get_keys"
    args = {"username": username,
            "session_uid": session_uid}

    response = service.call_function(function=function, args=args)

    public_key = None
    public_cert = None

    from Acquire.Crypto import PublicKey as _PublicKey
    from Acquire.ObjectStore import string_to_bytes as _string_to_bytes

    if "public_key" in response:
        public_key = _PublicKey.read_bytes(
                          _string_to_bytes(response["public_key"]))

    if "public_cert" in response:
        public_cert = _PublicKey.read_bytes(
                          _string_to_bytes(response["public_cert"]))

    return (public_key, public_cert)




__all__ = ["FileHandle"]


_magic_dict = {
    b"\x1f\x8b\x08": "gz",
    b"\x42\x5a\x68": "bz2",
    b"\x50\x4b\x03\x04": "zip"
    }


_max_magic_len = max(len(x) for x in _magic_dict)


def _should_compress(filename, filesize):
    """Return whether or not the passed file is worth compressing.
       It is not worth compressing very small files (<128 bytes) or
       already-compressed files
    """
    if filesize < 128:
        return False

    with open(filename, "rb") as FILE:
        file_start = FILE.read(_max_magic_len)

    for magic in _magic_dict.keys():
        if file_start.startswith(magic):
            return False

    return True


def _bz2compress(inputfile, outputfile=None):
    """Compress 'inputfile', writing the output to 'outputfile'
       If 'outputfile' is None, then this will create a new filename
       in the current directory for the file. This returns
       the filename for the compressed file
    """
    from Acquire.Client import compress as _compress
    return _compress(inputfile=inputfile, outputfile=outputfile,
                     compression_type="bz2")


class FileHandle:
    """This class holds all of the information about a file that is
       held in a Drive, including its size
       and checksum, and information about previous versions. It
       provides a handle that you can use to download or delete
       the file, to upload new versions, or to move the data between
       hot and cold storage or pay for extended storage
    """
    def __init__(self, filename=None, remote_filename=None,
                 aclrules=None, drive_uid=None, user_guid=None,
                 compress=True):
        """Construct a handle for the local file 'filename'. This will
           create the initial version of the file that can be uploaded
           to the storage service.
        """
        self._local_filename = None
        self._local_filedata = None
        self._compression = None
        self._compressed_filename = None
        self._drive_uid = drive_uid
        self._user_guid = None
        self._aclrules = None

        if filename is not None:
            self._user_guid = user_guid

            if aclrules is None:
                # will be automatically set to 'inherit' on the service
                self._aclrules = None
            else:
                from Acquire.Storage import create_aclrules as _create_aclrules
                self._aclrules = _create_aclrules(aclrules=aclrules)

            from Acquire.Access import get_filesize_and_checksum \
                as _get_filesize_and_checksum
            import os as _os

            (filesize, cksum) = _get_filesize_and_checksum(filename=filename)

            if compress and _should_compress(filename=filename,
                                             filesize=filesize):
                import bz2 as _bz2
                if filesize < 1048576:
                    # this is not big, so better to compress in memory
                    from Acquire.Access import get_size_and_checksum \
                        as _get_size_and_checksum
                    data = open(filename, "rb").read()
                    data = _bz2.compress(data)
                    (filesize, cksum) = _get_size_and_checksum(data=data)
                    self._local_filedata = data
                    self._compression = "bz2"
                else:
                    # this is a bigger file, so compress on disk
                    try:
                        self._compressed_filename = _bz2compress(
                                                        inputfile=filename)
                    except:
                        pass

                    if self._compressed_filename is not None:
                        self._compression = "bz2"
                        (filesize, cksum) = _get_filesize_and_checksum(
                                            filename=self._compressed_filename)
            elif filesize < 1048576:
                # this is small enough to hold in memory
                self._local_filedata = open(filename, "rb").read()

            if self._compressed_filename is None:
                self._local_filename = filename

            self._filesize = filesize
            self._checksum = cksum

            if remote_filename is None:
                self._filename = _os.path.split(filename)[1]
            else:
                self._filename = _os.path.split(remote_filename)[1]
        else:
            self._filename = None

    def __del__(self):
        """Ensure we delete the temporary file before being destroyed"""
        if self._compressed_filename is not None:
            import os as _os
            _os.unlink(self._compressed_filename)

    def __str__(self):
        """Return a string representation of the file"""
        if self.is_null():
            return "FileHandle::null"

        return "FileHandle(filename='%s')" % self.filename()

    def is_null(self):
        """Return whether or not this this null"""
        return self._filename is None

    def is_compressed(self):
        """Return whether or not the file is compressed on transport"""
        return self._compression is not None

    def compression_type(self):
        """Return a string describing the compression scheme used by the
           filehandle when transporting the file, or None if the data
           is not compressed
        """
        return self._compression

    def is_localdata(self):
        """Return whether or not this file is so small that the data
           is held in memory
        """
        return self._local_filedata is not None

    def local_filedata(self, uncompress=False):
        """Return the filedata for this file, assuming it is sufficiently
           small to be read in this way. Returns 'None' if not...

           If 'uncompress' is true, then uncompress the data
           (if it is compressed) before returning
        """
        if uncompress and self.is_compressed():
            if self._local_filedata is not None:
                import bz2 as _bz2
                return _bz2.decompress(self._local_filedata)
            else:
                return None
        else:
            return self._local_filedata

    def local_filename(self):
        """Return the local filename for this file"""
        if self.is_localdata():
            return None
        elif self.is_compressed():
            return self._compressed_filename
        else:
            return self._local_filename

    def drive_uid(self):
        """Return the UID of the drive on which this file is located"""
        return self._drive_uid

    def aclrules(self):
        """Return the ACL rules for this file"""
        return self._aclrules

    def user_guid(self):
        """Return the GUID of the user who has opened/uploaded this file"""
        return self._user_guid

    def filename(self):
        """Return the remote (object store) filename for this file"""
        return self._filename

    def filesize(self):
        """Return the size (in bytes) of this file"""
        if self.is_null():
            return 0
        else:
            return self._filesize

    def checksum(self):
        """Return the checksum of the contents of this file"""
        if self.is_null():
            return None
        else:
            return self._checksum

    def fingerprint(self):
        """Return a fingerprint for this file"""
        return "%s:%s:%s" % (self.filename(),
                             self.filesize(), self.checksum())

    def to_data(self):
        """Return a json-serialisable dictionary for this object. Note
           that this does not contain any information about the local
           file itself - just the name it should be called on the
           object store and the size, checksum and acl. If the file
           (or compressed file) is sufficiently small then this
           will also contain the packed version of that file data
        """
        data = {}

        if not self.is_null():
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            data["filename"] = self.filename()
            data["filesize"] = self.filesize()
            data["checksum"] = self.checksum()

            if self._aclrules is not None:
                data["aclrules"] = self._aclrules.to_data()

            data["drive_uid"] = self.drive_uid()
            data["user_guid"] = self.user_guid()

            if self._local_filedata is not None:
                from Acquire.ObjectStore import bytes_to_string \
                    as _bytes_to_string
                data["filedata"] = _bytes_to_string(self._local_filedata)

            if self._compression is not None:
                data["compression"] = self._compression

        return data

    @staticmethod
    def from_data(data):
        """Return an object created from the passed json-deserialised
           dictionary. Note that this does not contain any information
           about the local file itself - just the name it should be
           called on the object store and the size, checksum and acl.
           If the file (or compressed file) is sufficiently small then this
           will also contain the packed version of that file data
        """
        f = FileHandle()

        if data is not None and len(data) > 0:
            from Acquire.Storage import ACLRule as _ACLRule
            f._filename = data["filename"]
            f._filesize = int(data["filesize"])
            f._checksum = data["checksum"]
            f._drive_uid = data["drive_uid"]

            if "compression" in data:
                f._compression = data["compression"]

            if "aclrules" in data:
                from Acquire.Storage import ACLRules as _ACLRules
                f._aclrules = _ACLRules.from_data(data["aclrules"])

            if "filedata" in data:
                from Acquire.ObjectStore import string_to_bytes \
                    as _string_to_bytes

                f._local_filedata = _string_to_bytes(data["filedata"])

        return f




from enum import Enum as _Enum

__all__ = ["ACLRules", "ACLUserRules", "ACLGroupRules",
           "create_aclrules"]


class ACLRuleOperation(_Enum):
    MAX = "max"  # add rules together (most permissive)
    MIN = "min"  # add rules together (least permissive)
    SUB = "sub"  # subtract rules (why?)
    SET = "set"  # break - set first matching fully-resolved rule

    def to_data(self):
        return self.value

    def combine(self, acl1, acl2):
        if acl1 is None:
            return acl2
        elif acl2 is None:
            return acl1

        if self is ACLRuleOperation.SET:
            return acl1
        elif self is ACLRuleOperation.MAX:
            return acl1 + acl2
        elif self is ACLRuleOperation.MIN:
            return acl1 * acl2
        elif self is ACLRuleOperation.SUB:
            return acl1 - acl2
        else:
            return None

    @staticmethod
    def from_data(data):
        return ACLRuleOperation(data)


def create_aclrules(**kwargs):
    """Create an ACLRules object the passed set of rules, for example

        user_guid, aclrule  would set the aclrule for user to aclrule
        group_guid, aclrule would set the aclrule for group to aclrule

        default_rule  would set a default rule if nothing else matches
    """
    aclrules = None

    if "aclrules" in kwargs:
        aclrules = kwargs["aclrules"]

        if not isinstance(aclrules, ACLRules):
            aclrules = ACLRules(rule=aclrules)

        if "default" in kwargs:
            aclrules.set_default_rule(kwargs["default"])

    if "aclrule" in kwargs:
        from Acquire.Storage import ACLRule as _ACLRule

        aclrule = kwargs["aclrule"]
        if "user_guid" in kwargs:
            if aclrules is None:
                aclrules = ACLRules(default_rule=_ACLRule.denied())

            user_rules = ACLUserRules()
            user_rules.add_user_rule(kwargs["user_guid"], aclrule)
            aclrules.append(user_rules)
        elif "group_guid" in kwargs:
            if aclrules is None:
                aclrules = ACLRules(default_rule=_ACLRule.denied())

            group_rules = ACLGroupRules()
            group_rules.add_group_rule(kwargs["group_guid"], aclrule)
            aclrules.append(group_rules)
        else:
            if isinstance(aclrule, ACLRules):
                aclrules.append(aclrule)

    if "default" in kwargs:
        if aclrules is None:
            aclrules = ACLRules(default_rule=kwargs["default"])
        else:
            aclrules.set_default(kwargs["default"])

    if aclrules is None:
        aclrules = ACLRules()

    return aclrules


def _save_rule(rule):
    """Return a json-serialisable object for the passed rule"""
    return [rule.__class__.__name__, rule.to_data()]


def _load_rule(rule):
    """Return the rule loaded from the json-deserialised data"""
    try:
        classname = rule[0]
        classdata = rule[1]
    except:
        raise TypeError("Expected [classname, classdata]")

    if classname == "ACLRules":
        return ACLRules.from_data(classdata)
    elif classname == "ACLUserRules":
        return ACLUserRules.from_data(classdata)
    elif classname == "ACLGroupRules":
        return ACLGroupRules.from_data(classdata)
    elif classname == "ACLRule":
        from Acquire.Storage import ACLRule as _ACLRule
        return _ACLRule.from_data(classdata)
    else:
        raise TypeError("Unrecognised type '%s'" % classname)


class ACLGroupRules:
    """This class holds rules that apply to individual groups"""
    def __init__(self):
        """Construct, optionally starting with a default rule
           is no groups are matched
        """
        self._group_rules = {}

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __str__(self):
        s = []
        for group, rule in self._group_rules.items():
            s.append("%s => %s" % (group, rule))
        return "Group{%s}" % ", ".join(s)

    def resolve(self, must_resolve=True, **kwargs):
        """Resolve the rule for the user with specified group_guid.
           This returns None if there are no rules for this group
        """
        try:
            group_guid = kwargs["group_guid"]
        except:
            group_guid = None

        if group_guid in self._group_rules:
            rule = self._group_rules[group_guid]
            return rule.resolve(must_resolve=must_resolve, **kwargs)
        elif must_resolve:
            from Acquire.Storage import ACLRule as _ACLRule
            return _ACLRule.inherit().resolve(must_resolve=True, **kwargs)
        else:
            return None

    def add_group_rule(self, group_guid, rule):
        """Add a rule for the used with passed 'group_guid'"""
        self._group_rules[group_guid] = rule

    def to_data(self):
        """Return a json-serialisable representation of these rules"""
        data = {}

        for group_guid, rule in self._group_rules.items():
            data[group_guid] = _save_rule(rule)

        return data

    @staticmethod
    def from_data(data):
        """Return the rules constructed from the passed json-deserialised
           object
        """
        rules = ACLGroupRules()

        if data is not None and len(data) > 0:
            for group_guid, rule in data.items():
                rules.add_group_rule(group_guid, _load_rule(rule))

        return rules


class ACLUserRules:
    """This class holds rules that apply to individual users"""
    def __init__(self):
        """Construct, optionally starting with a default rule
           if no users are matched
        """
        self._user_rules = {}

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __str__(self):
        s = []
        for user, rule in self._user_rules.items():
            s.append("%s => %s" % (user, rule))
        return "User{%s}" % ", ".join(s)

    def resolve(self, must_resolve=True, **kwargs):
        """Resolve the rule for the user with specified user_guid.
           This returns None if there are no rules for this user
        """
        try:
            user_guid = kwargs["user_guid"]
        except:
            user_guid = None

        if user_guid in self._user_rules:
            rule = self._user_rules[user_guid]
            return rule.resolve(must_resolve=must_resolve, **kwargs)
        elif must_resolve:
            from Acquire.Storage import ACLRule as _ACLRule
            return _ACLRule.inherit().resolve(must_resolve=True, **kwargs)
        else:
            return None

    def add_user_rule(self, user_guid, rule):
        """Add a rule for the used with passed 'user_guid'"""
        self._user_rules[user_guid] = rule

    @staticmethod
    def _create(aclrule, user_guid, user_guids):
        rule = ACLUserRules()

        if user_guid is not None:
            rule.add_user_rule(user_guid, aclrule)

        if user_guids is not None:
            for user_guid in user_guids:
                rule.add_user_rule(user_guid, aclrule)

        return rule

    @staticmethod
    def owner(user_guid=None, user_guids=None):
        """Simple shorthand to create the rule that the specified
           user is the owner of the resource
        """
        from Acquire.Storage import ACLRule as _ACLRule
        return ACLUserRules._create(aclrule=_ACLRule.owner(),
                                    user_guid=user_guid,
                                    user_guids=user_guids)

    @staticmethod
    def executer(user_guid=None, user_guids=None):
        """Simple shorthand to create the rule that the specified
           user is the executer of the resource
        """
        from Acquire.Storage import ACLRule as _ACLRule
        return ACLUserRules._create(aclrule=_ACLRule.executer(),
                                    user_guid=user_guid,
                                    user_guids=user_guids)

    @staticmethod
    def writer(user_guid=None, user_guids=None):
        """Simple shorthand to create the rule that the specified
           user is the writer of the resource
        """
        from Acquire.Storage import ACLRule as _ACLRule
        return ACLUserRules._create(aclrule=_ACLRule.writer(),
                                    user_guid=user_guid,
                                    user_guids=user_guids)

    @staticmethod
    def reader(user_guid=None, user_guids=None):
        """Simple shorthand to create the rule that the specified
           user is the reader of the resource
        """
        from Acquire.Storage import ACLRule as _ACLRule
        return ACLUserRules._create(aclrule=_ACLRule.reader(),
                                    user_guid=user_guid,
                                    user_guids=user_guids)

    def to_data(self):
        """Return a json-serialisable representation of these rules"""
        data = {}

        for user_guid, rule in self._user_rules.items():
            data[user_guid] = _save_rule(rule)

        return data

    @staticmethod
    def from_data(data):
        """Return the rules constructed from the passed json-deserialised
           object
        """
        rules = ACLUserRules()

        if data is not None and len(data) > 0:
            for user_guid, rule in data.items():
                rules.add_user_rule(user_guid, _load_rule(rule))

        return rules


def _is_inherit(aclrule):
    """Return whether or not this passed rule is just an inherit-all"""
    from Acquire.Storage import ACLRule as _ACLRule

    if isinstance(aclrule, _ACLRule):
        if aclrule == _ACLRule.inherit():
            return True

    return False


class ACLRules:
    """This class holds a combination of ACL rules. These are parsed
       in order to get the ACL for a resource.

       By default, this is a simple inherit rule (meaning that
       it will inherit whatever comes from upstream)
    """
    def __init__(self, rule=None, rules=None, default_rule=None,
                 default_operation=ACLRuleOperation.MAX):
        """Construct, optionally starting with a default ACLRule
           for all users
        """
        if default_rule is None:
            self._is_simple_inherit = True
        elif _is_inherit(default_rule):
            self._is_simple_inherit = True
        else:
            self._is_simple_inherit = False
            self._default_rule = default_rule
            self._rules = []

        self.set_default_operation(default_operation)

        if rule is not None:
            self.prepend(rule)

        if rules is not None:
            for rule in rules:
                self.append(aclrule=rule[1], operation=rule[0])

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self.__dict__ == other.__dict__
        else:
            return False

    def __str__(self):
        if self._is_simple_inherit:
            return "inherit"

        s = []
        for rule in rules:
            s.append("%s" % rule)

        if self._default_rule is not None:
            s.append("DEFAULT %s" % self._default_rule)

        return s

    def is_simple_inherit(self):
        """Return whether or not this set of rules is a simple
           'inherit all'
        """
        return self._is_simple_inherit

    def set_default_operation(self, default_operation):
        """Set the default operation used to combine together rules"""
        if not isinstance(default_operation, ACLRuleOperation):
            raise TypeError(
                "The default operation must be type ACLRuleOperation")

        self._default_operation = default_operation

    def set_default_rule(self, aclrule):
        """Set the default rule if nothing else matches (optionally
           also specifying the default operation to combine rules)
        """
        if self._is_simple_inherit:
            if _is_inherit(aclrule):
                return
            else:
                self._is_simple_inherit = False
                self._default_rule = aclrule
                self._default_operation = ACLRuleOperation.MAX
                self._rules = []
        else:
            self._default_rule = aclrule

    def append(self, aclrule, operation=None, ensure_owner=False):
        """Append a rule onto the set of rules. This will resolve any
           conflicts in the rules. If 'ensure_owner' is True, then
           this will ensure that there is at least one user who
           has unambiguous ownership of the resource controlled by
           these ACL rules
        """
        try:
            idx = len(self._rules)
        except:
            idx = 2

        self.insert(idx=idx, aclrule=aclrule,
                    operation=operation, ensure_owner=ensure_owner)

    def prepend(self, aclrule, operation=None, ensure_owner=False):
        """Prepend a rule onto the set of rules"""
        self.insert(idx=0, aclrule=aclrule,
                    operation=operation, ensure_owner=ensure_owner)

    def insert(self, idx, aclrule, operation=None, ensure_owner=False):
        """Insert the passed rule at index 'idx', specifying the operation
           used to combine this rule with what has gone before
           (defaults to self._default_operation)
        """
        if operation is not None:
            if not isinstance(operation, ACLRuleOperation):
                raise TypeError(
                    "The ACL operation must be type ACLRuleOperation")

        if self._is_simple_inherit:
            if _is_inherit(aclrule):
                return
            else:
                from Acquire.Storage import ACLRule as _ACLRule
                self._is_simple_inherit = False
                self._default_rule = None
                self._rules = [_ACLRule.inherit()]

        if operation is not None:
            self._rules.insert(idx, (operation, aclrule))
        else:
            self._rules.insert(idx, aclrule)

        if ensure_owner:
            # need to write code to ensure there is at least one owner
            pass

    def rules(self):
        """Return the list of ACL rules that will be applied
           in order (including the default rule, if set)
        """
        if self._is_simple_inherit:
            from Acquire.Storage import ACLRule as _ACLRule
            return [(self._default_operation, _ACLRule.inherit())]
        else:
            import copy as copy
            r = copy.copy(self._rules)
            if self._default_rule is not None:
                r.append((self._default_operation, self._default_rule))

            return r

    def resolve(self, must_resolve=True, **kwargs):
        """Resolve the rule based on the passed kwargs. This will
           resolve the rules in order the final ACLRule has been
           generated. If 'must_resolve' is True, then
           this is guaranteed to return a fully-resolved simple ACLRule
        """
        from Acquire.Storage import ACLRule as _ACLRule

        if self._is_simple_inherit:
            return _ACLRule.inherit().resolve(must_resolve=must_resolve,
                                              **kwargs)

        result = None
        must_break = False

        for rule in self._rules:
            if isinstance(rule, tuple):
                op = rule[0]
                rule = rule[1]
            else:
                op = self._default_operation

            # resolve the rule...
            rule = rule.resolve(must_resolve=False, **kwargs)

            if rule is not None:
                if op is ACLRuleOperation.SET:
                    # take the first matching rule
                    result = rule
                    must_break = True
                    break
                elif result is None:
                    result = rule
                else:
                    result = op.combine(result, rule)

        if (not must_break) and (self._default_rule is not None):
            rule = self._default_rule.resolve(must_resolve=False, **kwargs)
            if result is None:
                result = rule
            else:
                result = self._default_operation.combine(result, rule)

        # should now have a fully resolved ACLRule...
        if result is None:
            return _ACLRule.denied()

        if not isinstance(result, _ACLRule):
            raise PermissionError(
                "Did not fully resolve the ACLRule - got %s" % str(result))

        if not result.is_fully_resolved():
            result = result.resolve(must_resolve=True, **kwargs)

        # we have not been able to generate a fully-resolved ACL
        return result

    @staticmethod
    def from_data(data):
        """Construct these rules from the passed json-serialised
           dictionary
        """
        if isinstance(data, str) and data == "inherit":
                return ACLRules()

        if data is not None and len(data) > 0:

            if "default_rule" in data:
                default_rule = _load_rule(data["default_rule"])
            else:
                default_rule = None

            if "default_operation" in data:
                default_operation = \
                    ACLRuleOperation.from_data(data["default_operation"])
            else:
                default_operation = ACLRuleOperation.MAX

            if "rules" in data:
                rules = []
                for rule in data["rules"]:
                    if isinstance(rule, tuple):
                        rules.append((ACLRuleOperation.from_data(rule[0]),
                                      _load_rule(rule[1])))
                    else:
                        rules.append((None, _load_rule(rule)))
            else:
                rules = None

            return ACLRules(default_rule=default_rule,
                            default_operation=default_operation,
                            rules=rules)
        else:
            return ACLRules()

    def to_data(self):
        """Return a json-serialisable dictionary of these rules"""
        if self._is_simple_inherit:
            return "inherit"

        data = {}

        if len(self._rules) > 0:
            rules = []

            for rule in self._rules:
                if isinstance(rule, tuple):
                    rules.append((rule[0].to_data(), _save_rule(rule[1])))
                else:
                    rules.append(_save_rule(rule))

            data["rules"] = rules

        if self._default_rule is not None:
            data["default_rule"] = _save_rule(self._default_rule)

        if self._default_operation is not ACLRuleOperation.MAX:
            data["default_operation"] = self._default_operation.to_data()

        return data




__all__ = ["Drive"]


def _get_storage_url():
    """Function to discover and return the default storage url"""
    return "http://fn.acquire-aaai.com:8080/t/storage"


def _get_storage_service(storage_url=None):
    """Function to return the storage service for the system"""
    if storage_url is None:
        storage_url = _get_storage_url()

    from Acquire.Client import Service as _Service
    service = _Service(storage_url)

    if not service.is_storage_service():
        from Acquire.Client import LoginError
        raise LoginError(
            "You can only use a valid storage service to get CloudDrive info! "
            "The service at '%s' is a '%s'" %
            (storage_url, service.service_type()))

    if service.service_url() != storage_url:
        service.update_service_url(storage_url)

    return service


def _create_drive(user, name, drivemeta, storage_service):
    """Internal function used to create a Drive"""
    drive = Drive()
    drive._name = drivemeta.name()
    drive._drive_uid = drivemeta.uid()
    drive._container = drivemeta.container_uids()
    drive._acl = drivemeta.acl()
    drive._aclrules = drivemeta.aclrules()
    drive._user = user
    drive._storage_service = storage_service
    return drive


def _get_drive(user, name=None, storage_service=None, storage_url=None,
               autocreate=True):
    """Return the drive called 'name' of the passed user. Note that the
       user must be authenticated to call this function. The name
       will default to 'main' if it is not set, and the drive will
       be created automatically is 'autocreate' is True and the
       drive does not exist
    """
    if storage_service is None:
        storage_service = _get_storage_service(storage_url)
    else:
        if not storage_service.is_storage_service():
            raise TypeError("You can only query drives using "
                            "a valid storage service")

    if name is None:
        name = "main"
    else:
        name = str(name)

    if autocreate:
        autocreate = True
    else:
        autocreate = False

    from Acquire.Client import Authorisation as _Authorisation
    authorisation = _Authorisation(resource="UserDrives", user=user)

    args = {"authorisation": authorisation.to_data(),
            "name": name, "autocreate": autocreate}

    response = storage_service.call_function(function="open_drive", args=args)

    from Acquire.Client import DriveMeta as _DriveMeta

    return _create_drive(user=user, name=name, storage_service=storage_service,
                         drivemeta=_DriveMeta.from_data(response["drive"]))


class Drive:
    """This class provides a handle to a user's drive (space
       to hold files and folders). A drive is associated with
       a single storage service and can be shared amongst several
       users. Each drive has a unique UID, with users assiging
       their own shorthand names.
    """
    def __init__(self, user=None, name=None, storage_service=None,
                 storage_url=None, autocreate=True):
        """Construct a handle to the drive that the passed user
           calls 'name' on the passed storage service. If
           'autocreate' is True and the user is logged in then
           this will automatically create the drive if
           it doesn't exist already
        """
        if user is not None:
            drive = _get_drive(user=user, name=name,
                               storage_service=storage_service,
                               storage_url=storage_url, autocreate=autocreate)

            from copy import copy as _copy
            self.__dict__ = _copy(drive.__dict__)
        else:
            self._drive_uid = None

    def __str__(self):
        if self.is_null():
            return "Drive::null"
        else:
            return "Drive(user='%s', name='%s')" % \
                    (self._user.username(), self.name())

    def is_null(self):
        """Return whether or not this drive is null"""
        return self._drive_uid is None

    def acl(self):
        """Return the access control list for the user on this drive.
           If the ACL is not known, then None is returned
        """
        try:
            return self._acl
        except:
            return None

    def aclrules(self):
        """Return the ACL rules used to grant access to this drive. This
           is only visible to owners of this drive. If it is not visible,
           then None is returned
        """
        try:
            return self._aclrules
        except:
            return None

    def name(self):
        """Return the name given to this drive by the user"""
        return self._name

    def uid(self):
        """Return the UID of this drive"""
        return self._drive_uid

    def guid(self):
        """Return the global UID of this drive (combination of the
           UID of the storage service and UID of the drive)
        """
        if self.is_null():
            return None
        else:
            return "%s@%s" % (self.storage_service().uid(), self.uid())

    def storage_service(self):
        """Return the storage service for this drive"""
        if self.is_null():
            return None
        else:
            return self._storage_service

    def upload(self, filename, uploaded_name=None, aclrules=None):
        """Upload the file at 'filename' to this drive, assuming we have
           write access to this drive. The local file 'filename' will be
           uploaded to the drive as the file called 'filename' (just the
           filename - not the full path). If you want to specify the
           uploaded name then set this as "uploaded_name" (which again will
           just be a filename - no paths). If a file with this name exists,
           then this will upload a new version (assuming you have permission).
           Otherwise this will create a new file. You can set the
           ACL rules used to grant access to this file via 'aclrule'.
           If this is not set, then the rules will be derived from either
           the last version of the file, or inherited from the drive.
        """
        if self.is_null():
            raise PermissionError("Cannot upload a file to a null drive!")

        if uploaded_name is None:
            uploaded_name = filename

        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.Client import FileHandle as _FileHandle
        from Acquire.Client import PAR as _PAR
        from Acquire.Client import FileMeta as _FileMeta

        filehandle = _FileHandle(filename=filename,
                                 remote_filename=uploaded_name,
                                 drive_uid=self.uid(),
                                 user_guid=self._user.guid(),
                                 aclrules=aclrules)

        authorisation = _Authorisation(
                            resource="upload %s" % filehandle.fingerprint(),
                            user=self._user)

        args = {"filehandle": filehandle.to_data(),
                "authorisation": authorisation.to_data()}

        if not filehandle.is_localdata():
            # we will need to upload against a PAR, so need to tell
            # the service how to encrypt the PAR...
            privkey = self._user.session_key()
            args["encryption_key"] = privkey.public_key().to_data()

        # will eventually need to authorise payment...

        response = self.storage_service().call_function(
                                function="upload", args=args)

        filemeta = _FileMeta.from_data(response["filemeta"])

        # if this was a large file, then we will receive a PAR back
        # which must be used to upload the file
        if not filehandle.is_localdata():
            par = _PAR.from_data(response["upload_par"])
            par.write(privkey).set_object_from_file(
                                    filehandle.local_filename())

            authorisation = _Authorisation(
                                resource="uploaded %s" % par.uid(),
                                user=self._user)

            args = {"drive_uid": self.uid(),
                    "authorisation": authorisation.to_data(),
                    "par_uid": par.uid()}

            self.storage_service().call_function(
                                       function="uploaded", args=args)

        return filemeta

    def download(self, filename, downloaded_name=None, version=None,
                 dir=None):
        """Download the file called 'filename' from this drive into
           the local directory, or 'dir' if specified,
           ideally called 'filename'
           (or 'downloaded_name' if that is specified). If a local
           file exists with this name, then a new, unique filename
           will be used. This returns a dictionary mapping the
           downloaded filename to the FileMeta of the file

           Note that this only downloads files for which you
           have read-access. If the file is not readable then
           an exception is raised and nothing is returned

           If 'version' is specified then download a specific version
           of the file. Otherwise download the latest version
        """
        if self.is_null():
            raise PermissionError("Cannot download from a null drive!")

        if downloaded_name is None:
            downloaded_name = filename

        from Acquire.Client import create_new_file as \
            _create_new_file

        downloaded_name = _create_new_file(filename=downloaded_name,
                                           dir=dir)

        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.Client import FileMeta as _FileMeta

        authorisation = _Authorisation(
                            resource="download %s %s" % (self.uid(),
                                                         filename),
                            user=self._user)

        privkey = self._user.session_key()

        args = {"drive_uid": self.uid(),
                "filename": filename,
                "authorisation": authorisation.to_data(),
                "encryption_key": privkey.public_key().to_data()}

        if version is not None:
            from Acquire.ObjectStore import datetime_to_string \
                as _datetime_to_string
            args["version"] = _datetime_to_string(version)

        response = self.storage_service().call_function(
                                function="download", args=args)

        filemeta = _FileMeta.from_data(response["filemeta"])

        if "filedata" in response:
            # we have already downloaded the file to 'filedata'
            filedata = response["filedata"]

            from Acquire.ObjectStore import string_to_bytes \
                as _string_to_bytes
            filedata = _string_to_bytes(response["filedata"])
            del response["filedata"]

            # validate that the size and checksum are correct
            filemeta.assert_correct_data(filedata)

            if filemeta.is_compressed():
                # uncompress the data
                from Acquire.Client import uncompress as _uncompress
                filedata = _uncompress(
                                inputdata=filedata,
                                compression_type=filemeta.compression_type())

            # write the data to the specified local file...
            with open(downloaded_name, "wb") as FILE:
                FILE.write(filedata)
                FILE.flush()
        else:
            from Acquire.Client import PAR as _PAR
            par = _PAR.from_data(response["upload_par"])
            par.read(privkey).get_object_as_file(downloaded_name)

            # validate that the size and checksum are correct
            filemeta.assert_correct_data(filename=downloaded_name)

            # uncompress the file if desired
            if filemeta.is_compressed():
                from Acquire.Client import uncompress as _uncompress
                _uncompress(inputfile=downloaded_name,
                            outputfile=downloaded_name,
                            compression_type=filemeta.compression_type())

            # everything is ok, so we don't need the PAR any more
            authorisation = _Authorisation(
                                resource="downloaded %s" % par.uid(),
                                user=self._user)

            args = {"drive_uid": self.uid(),
                    "authorisation": authorisation.to_data(),
                    "par_uid": par.uid()}

            self.storage_service().call_function(
                                       function="downloaded", args=args)

        filemeta._set_drive(self)

        return (downloaded_name, filemeta)

    @staticmethod
    def _list_drives(user, drive_uid=None,
                     storage_service=None, storage_url=None):
        """Return a list of all of the DriveMetas of the drives accessible
           at the top-level by the passed user on the passed storage
           service
        """
        if storage_service is None:
            storage_service = _get_storage_service(storage_url)
        else:
            if not storage_service.is_storage_service():
                raise TypeError("You can only query drives using "
                                "a valid storage service")

        from Acquire.Client import Authorisation as _Authorisation
        authorisation = _Authorisation(resource="UserDrives", user=user)

        args = {"authorisation": authorisation.to_data()}

        if drive_uid is not None:
            args["drive_uid"] = str(drive_uid)

        response = storage_service.call_function(
                                    function="list_drives", args=args)

        from Acquire.ObjectStore import string_to_list as _string_to_list
        from Acquire.Client import DriveMeta as _DriveMeta

        return _string_to_list(response["drives"], _DriveMeta)

    @staticmethod
    def list_toplevel_drives(user, storage_service=None, storage_url=None):
        """Return a list of all of the DriveMetas of the drives accessible
           at the top-level by the passed user on the passed storage
           service
        """
        return Drive._list_drives(user=user,
                                  storage_service=storage_service,
                                  storage_url=storage_url)

    def list_drives(self):
        """Return a list of the DriveMetas of all of the drives contained
           in this drive that are accessible to the user
        """
        if self.is_null():
            return []
        else:
            return Drive._list_drives(user=self._user,
                                      drive_uid=self._drive_uid,
                                      storage_service=self._storage_service)

    def list_files(self, include_metadata=False):
        """Return a list of the FileMetas of all of the files contained
           in this drive
        """
        if self.is_null():
            return []

        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.ObjectStore import string_to_list as _string_to_list
        from Acquire.Storage import FileMeta as _FileMeta

        if include_metadata:
            include_metadata = True
        else:
            include_metadata = False

        authorisation = _Authorisation(resource="list_files",
                                       user=self._user)

        args = {"authorisation": authorisation.to_data(),
                "drive_uid": self._drive_uid,
                "include_metadata": include_metadata}

        response = self.storage_service().call_function(function="list_files",
                                                        args=args)

        files = _string_to_list(response["files"], _FileMeta)

        for f in files:
            f._set_drive(self)

        return files

    def list_versions(self, filename, include_metadata=False):
        """Return a list of all of the versions of the specified file.
           This returns an empty list if there are no versions of this
           file
        """
        if self.is_null():
            return []

        if include_metadata:
            include_metadata = True
        else:
            include_metadata = False

        from Acquire.Client import Authorisation as _Authorisation

        authorisation = _Authorisation(resource="list_versions %s" % filename,
                                       user=self._user)

        args = {"authorisation": authorisation.to_data(),
                "drive_uid": self._drive_uid,
                "include_metadata": include_metadata,
                "filename": filename}

        response = self.storage_service().call_function(
                                                function="list_versions",
                                                args=args)

        from Acquire.ObjectStore import string_to_list \
            as _string_to_list
        from Acquire.Storage import FileMeta as _FileMeta

        versions = _string_to_list(response["versions"], _FileMeta)
        for version in versions:
            version._set_drive(self)

        return versions




import os as _os
import shutil as _shutil
import datetime as _datetime
import uuid as _uuid
import json as _json
import glob as _glob
import threading
import uuid as _uuid

_rlock = threading.RLock()

__all__ = ["Testing_ObjectStore"]


class Testing_ObjectStore:
    """This is a dummy object store that writes objects to
       the standard posix filesystem when running tests
    """
    @staticmethod
    def create_bucket(bucket, bucket_name, compartment=None):
        """Create and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. This will raise an
           ObjectStoreError if this bucket already exists
        """
        bucket_name = str(bucket_name)

        if compartment is not None:
            if compartment.endswith("/"):
                bucket = compartment
            else:
                bucket = "%s/" % compartment

        full_name = _os.path.join(_os.path.split(bucket)[0], bucket_name)

        if _os.path.exists(full_name):
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError(
                "CANNOT CREATE NEW BUCKET '%s': EXISTS!" % bucket_name)

        _os.makedirs(full_name)

        return full_name

    @staticmethod
    def get_bucket(bucket, bucket_name, compartment=None,
                   create_if_needed=True):
        """Find and return a new bucket in the object store called
           'bucket_name', optionally placing it into the compartment
           identified by 'compartment'. If 'create_if_needed' is True
           then the bucket will be created if it doesn't exist. Otherwise,
           if the bucket does not exist then an exception will be raised.
        """
        bucket_name = str(bucket_name)

        if compartment is not None:
            if compartment.endswith("/"):
                bucket = compartment
            else:
                bucket = "%s/" % compartment

        full_name = _os.path.join(_os.path.split(bucket)[0], bucket_name)

        if not _os.path.exists(full_name):
            if create_if_needed:
                _os.makedirs(full_name)
            else:
                from Acquire.ObjectStore import ObjectStoreError
                raise ObjectStoreError(
                    "There is no bucket available called '%s' in "
                    "compartment '%s'" % (bucket_name, compartment))

        return full_name

    @staticmethod
    def create_par(bucket, encrypt_key, key=None, readable=True,
                   writeable=False, duration=3600):
        """Create a pre-authenticated request for the passed bucket and
           key (if key is None then the request is for the entire bucket).
           This will return a PAR object that will contain a URL that can
           be used to access the object/bucket. If writeable is true, then
           the URL will also allow the object/bucket to be written to.
           PARs are time-limited. Set the lifetime in seconds by passing
           in 'duration' (by default this is one hour). Note that you must
           pass in a public key that will be used to encrypt this PAR. This is
           necessary as the PAR grants access to anyone who can decrypt
           the URL
        """
        from Acquire.Crypto import PublicKey as _PublicKey

        if not isinstance(encrypt_key, _PublicKey):
            from Acquire.Client import PARError
            raise PARError(
                "You must supply a valid PublicKey to encrypt the "
                "returned PAR")

        if key is not None:
            if not _os.path.exists("%s/%s._data" % (bucket, key)):
                from Acquire.Client import PARError
                raise PARError(
                    "The object '%s' in bucket '%s' does not exist!" %
                    (key, bucket))
        elif not _os.path.exists(bucket):
            from Acquire.Client import PARError
            raise PARError("The bucket '%s' does not exist!" % bucket)

        url = "file://%s" % bucket

        if key:
            url = "%s/%s" % (url, key)

        # get the time this PAR was created
        from Acquire.ObjectStore import get_datetime_now as _get_datetime_now
        created_datetime = _get_datetime_now()

        # get the UTC datetime when this PAR should expire
        expires_datetime = created_datetime + \
            _datetime.timedelta(seconds=duration)

        # mimic limitations of OCI - cannot have a bucket PAR with
        # read permissions!
        if (key is None) and readable:
            from Acquire.Client import PARError
            raise PARError(
                "You cannot create a Bucket PAR that has read permissions "
                "due to a limitation in the underlying platform")

        from Acquire.Client import PAR as _PAR

        return _PAR(url=url, key=key, encrypt_key=encrypt_key,
                    created_datetime=created_datetime,
                    expires_datetime=expires_datetime,
                    is_readable=readable, is_writeable=writeable,
                    par_id=str(_uuid.uuid4()),
                    driver="testing_objstore")

    @staticmethod
    def delete_par(bucket, par):
        """Delete the passed PAR, which provides access to data in the
           passed bucket
        """
        from Acquire.Client import PAR as _PAR

        if not isinstance(par, _PAR):
            raise TypeError("The PAR must be of type PAR")

        if par.driver() != "testing_objstore":
            raise ValueError("Cannot delete a PAR that was not created "
                             "by the testing object store")

    @staticmethod
    def get_object_as_file(bucket, key, filename):
        """Get the object contained in the key 'key' in the passed 'bucket'
           and writing this to the file called 'filename'"""

        if not _os.path.exists("%s/%s._data" % (bucket, key)):
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError("No object at key '%s'" % key)

        _shutil.copy("%s/%s._data" % (bucket, key), filename)

    @staticmethod
    def get_object(bucket, key):
        """Return the binary data contained in the key 'key' in the
           passed bucket"""

        with _rlock:
            if _os.path.exists("%s/%s._data" % (bucket, key)):
                return open("%s/%s._data" % (bucket, key), "rb").read()
            else:
                from Acquire.ObjectStore import ObjectStoreError
                raise ObjectStoreError("No object at key '%s'" % key)

    @staticmethod
    def get_string_object(bucket, key):
        """Return the string in 'bucket' associated with 'key'"""
        return Testing_ObjectStore.get_object(bucket, key).decode("utf-8")

    @staticmethod
    def get_object_from_json(bucket, key):
        """Return an object constructed from json stored at 'key' in
           the passed bucket. This returns None if there is no data
           at this key
        """

        data = None

        try:
            data = Testing_ObjectStore.get_string_object(bucket, key)
        except:
            return None

        return _json.loads(data)

    @staticmethod
    def get_all_object_names(bucket, prefix=None):
        """Returns the names of all objects in the passed bucket"""

        root = bucket

        if prefix is not None:
            root = "%s/%s" % (bucket, prefix)

        root_len = len(bucket) + 1

        subdir_names = _glob.glob("%s*" % root)

        object_names = []

        while True:
            names = subdir_names
            subdir_names = []

            for name in names:
                if name.endswith("._data"):
                    # remove the  ._data at the end
                    name = name[root_len:-6]
                    while name.endswith("/"):
                        name = name[0:-1]

                    while name.startswith("/"):
                        name = name[1:]

                    if len(name) > 0:
                        object_names.append(name)
                elif _os.path.isdir(name):
                    subdir_names += _glob.glob("%s/*" % name)

            if len(subdir_names) == 0:
                break

        return object_names

    @staticmethod
    def get_all_objects(bucket, prefix=None):
        """Return all of the objects in the passed bucket"""
        objects = {}
        names = Testing_ObjectStore.get_all_object_names(bucket, prefix)

        for name in names:
            objects[name] = Testing_ObjectStore.get_object(bucket, name)

        return objects

    @staticmethod
    def get_all_objects_from_json(bucket, prefix=None):
        """Return all of the json objects in the passed bucket as
           json-deserialised objects
        """
        objects = Testing_ObjectStore.get_all_objects(bucket, prefix)

        names = list(objects.keys())

        for name in names:
            try:
                s = objects[name].decode("utf-8")
                objects[name] = _json.loads(s)
            except:
                del objects[name]

        return objects

    @staticmethod
    def get_all_strings(bucket, prefix=None):
        """Return all of the strings in the passed bucket"""
        objects = Testing_ObjectStore.get_all_objects(bucket, prefix)

        names = list(objects.keys())

        for name in names:
            try:
                s = objects[name].decode("utf-8")
                objects[name] = s
            except:
                del objects[name]

        return objects

    @staticmethod
    def set_object(bucket, key, data):
        """Set the value of 'key' in 'bucket' to binary 'data'"""

        filename = "%s/%s._data" % (bucket, key)

        with _rlock:
            try:
                with open(filename, 'wb') as FILE:
                    FILE.write(data)
                    FILE.flush()
            except:
                dir = "/".join(filename.split("/")[0:-1])
                _os.makedirs(dir, exist_ok=True)
                with open(filename, 'wb') as FILE:
                    FILE.write(data)
                    FILE.flush()

    @staticmethod
    def set_object_from_file(bucket, key, filename):
        """Set the value of 'key' in 'bucket' to equal the contents
           of the file located by 'filename'"""

        Testing_ObjectStore.set_object(bucket, key,
                                       open(filename, 'rb').read())

    @staticmethod
    def set_string_object(bucket, key, string_data):
        """Set the value of 'key' in 'bucket' to the string 'string_data'"""
        Testing_ObjectStore.set_object(bucket, key,
                                       string_data.encode("utf-8"))

    @staticmethod
    def set_object_from_json(bucket, key, data):
        """Set the value of 'key' in 'bucket' to equal to contents
           of 'data', which has been encoded to json"""
        Testing_ObjectStore.set_string_object(bucket, key, _json.dumps(data))

    @staticmethod
    def delete_all_objects(bucket, prefix=None):
        """Deletes all objects..."""
        if prefix:
            _shutil.rmtree("%s/%s" % (bucket, prefix), ignore_errors=True)
        else:
            _shutil.rmtree(bucket, ignore_errors=True)

    @staticmethod
    def delete_object(bucket, key):
        """Removes the object at 'key'"""
        try:
            _os.remove("%s/%s._data" % (bucket, key))
        except:
            pass

    @staticmethod
    def clear_all_except(bucket, keys):
        """Removes all objects from the passed 'bucket' except those
           whose keys are or start with any key in 'keys'"""

        names = Testing_ObjectStore.get_all_object_names(bucket)

        for name in names:
            remove = True

            for key in keys:
                if name.startswith(key):
                    remove = False
                    break

            if remove:
                Testing_ObjectStore.delete_object(bucket, key)

    @staticmethod
    def get_size_and_checksum(bucket, key):
        """Return the object size (in bytes) and checksum of the
           object in the passed bucket at the specified key
        """
        filepath = "%s/%s._data" % (bucket, key)

        if not _os.path.exists(filepath):
            from Acquire.ObjectStore import ObjectStoreError
            raise ObjectStoreError("No object at key '%s'" % key)

        from Acquire.Access import get_filesize_and_checksum \
            as _get_filesize_and_checksum

        return _get_filesize_and_checksum(filepath)




from Acquire.Service import ServiceError

__all__ = ["IdentityServiceError", "LoginSessionError", "UsernameError",
           "ExistingAccountError", "UserValidationError",
           "AuthorisationError"]


class IdentityServiceError(ServiceError):
    pass


class LoginSessionError(Exception):
    pass


class UsernameError(Exception):
    pass


class ExistingAccountError(Exception):
    pass


class UserValidationError(Exception):
    pass


class AuthorisationError(Exception):
    pass




__all__ = ["DriveInfo"]

_drive_root = "storage/drive"

_fileinfo_root = "storage/file"

_par_root = "storage/file_pars"


class DriveInfo:
    """This class provides a service-side handle to the information
       about a particular cloud drive
    """
    def __init__(self, drive_uid=None, user_guid=None,
                 is_authorised=False, parent_drive_uid=None):
        """Construct a DriveInfo for the drive with UID 'drive_uid',
           and optionally the GUID of the user making the request
           (and whether this was authorised). If this drive
           has a parent then it is a sub-drive and not recorded
           in the list of top-level drives
        """
        self._drive_uid = drive_uid
        self._parent_drive_uid = parent_drive_uid
        self._user_guid = user_guid
        self._is_authorised = is_authorised

        if self._drive_uid is not None:
            self.load()

    def __str__(self):
        if self.is_null():
            return "Drive::null"
        else:
            return "Drive(%s)" % self._drive_uid

    def is_null(self):
        return self._drive_uid is None

    def _drive_key(self):
        """Return the key for this drive in the object store"""
        return "%s/%s/info" % (_drive_root, self._drive_uid)

    def uid(self):
        """Return the UID of this drive"""
        return self._drive_uid

    def _get_metadata_bucket(self):
        """Return the bucket that contains all of the metadata about
           the files for this drive
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.Service import get_this_service as _get_this_service

        service = _get_this_service()
        bucket = _get_service_account_bucket()
        bucket_name = "user_metadata"

        try:
            return _ObjectStore.get_bucket(
                            bucket=bucket, bucket_name=bucket_name,
                            compartment=service.storage_compartment(),
                            create_if_needed=True)
        except Exception as e:
            from Acquire.ObjectStore import RequestBucketError
            raise RequestBucketError(
                "Unable to open the bucket '%s': %s" % (bucket_name, str(e)))

    def _get_file_bucket(self):
        """Return the bucket that contains all of the files for this
           drive
        """
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.Service import get_this_service as _get_this_service

        service = _get_this_service()
        bucket = _get_service_account_bucket()
        bucket_name = "user_files"

        try:
            return _ObjectStore.get_bucket(
                            bucket=bucket, bucket_name=bucket_name,
                            compartment=service.storage_compartment(),
                            create_if_needed=True)
        except Exception as e:
            from Acquire.ObjectStore import RequestBucketError
            raise RequestBucketError(
                "Unable to open the bucket '%s': %s" % (bucket_name, str(e)))

    def upload_complete(self, par_uid, authorisation):
        """Call this function to signify that the file associated with
           the PAR with UID 'par_uid' has been uploaded (must have matching
           authorisation)
        """
        from Acquire.Client import PAR as _PAR
        from Acquire.Client import Authorisation as _Authorisation
        from Acquire.Client import FileMeta as _FileMeta

        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The authorisation must be of type Authorisation")

        authorisation.verify("uploaded %s" % par_uid)

        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        par_key = "%s/%s" % (_par_root, par_uid)
        file_bucket = self._get_file_bucket()
        metadata_bucket = self._get_metadata_bucket()

        data = _ObjectStore.get_object_from_json(metadata_bucket, par_key)

        par = _PAR.from_data(data["par"])
        file_key = data["file_key"]
        user_guid = data["user_guid"]
        expected_objsize = data["filesize"]
        expected_checksum = data["checksum"]

        # the same user who requested the PAR must signify that the upload
        # has completed
        if authorisation.user_guid() != user_guid:
            raise PermissionError(
                "Only the user with GUID %s can signify that the "
                "upload is complete. Not user with GUID %s" %
                (user_guid, authorisation.user_guid()))

        # check that the file uploaded matches what was promised
        (objsize, checksum) = _ObjectStore.get_size_and_checksum(file_bucket,
                                                                 file_key)

        if expected_objsize != objsize or expected_checksum != checksum:
            from Acquire.Storage import FileValidationError
            raise FileValidationError(
                "The file uploaded does not match what was promised. "
                "size: %s versus %s, checksum: %s versus %s. Please try "
                "to upload the file again." %
                (expected_objsize, objsize,
                 expected_checksum, checksum))

            # probably should delete the broken object here...

        # SHOULD HERE RECEIPT THE STORAGE TRANSACTION

        _ObjectStore.delete_par(bucket=file_bucket, par=par)
        _ObjectStore.delete_object(bucket=metadata_bucket, key=par_key)

    def upload(self, filehandle, authorisation, encrypt_key=None):
        """Upload the file associated with the passed filehandle.
           If the filehandle has the data embedded, then this uploads
           the file data directly and returns a FileMeta for the
           result. Otherwise, this returns a PAR which should
           be used to upload the data. The PAR will be encrypted
           using 'encrypt_key'
        """
        from Acquire.Client import FileHandle as _FileHandle
        from Acquire.Storage import FileInfo as _FileInfo
        from Acquire.Identity import Authorisation as _Authorisation
        from Acquire.Crypto import PublicKey as _PublicKey
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded

        if not isinstance(filehandle, _FileHandle):
            raise TypeError("The fileinfo must be of type FileInfo")

        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The authorisation must be of type Authorisation")

        if encrypt_key is not None:
            if not isinstance(encrypt_key, _PublicKey):
                raise TypeError("The encryption key must be of type PublicKey")

        authorisation.verify("upload %s" % filehandle.fingerprint())

        user_guid = authorisation.user_guid()

        drive_acl = self.aclrules().resolve(user_guid=user_guid)

        if not drive_acl.is_writeable():
            raise PermissionError(
                "You do not have permission to write to this drive. "
                "Your permissions are %s" % str(drive_acl))

        # now generate a FileInfo for this FileHandle
        fileinfo = _FileInfo(drive_uid=self._drive_uid,
                             filehandle=filehandle,
                             user_guid=authorisation.user_guid())

        # resolve the ACL for the file from this FileHandle
        file_acl = fileinfo.aclrules().resolve(upstream=drive_acl,
                                               user_guid=user_guid)

        if not file_acl.is_writeable():
            raise PermissionError(
                "Despite having write permission to the drive, you "
                "do not have write permission for the file. Your file "
                "permissions are %s" % str(file_acl))

        file_bucket = self._get_file_bucket()
        metadata_bucket = self._get_metadata_bucket()

        file_key = fileinfo.latest_version()._file_key()

        filedata = None

        if filehandle.is_localdata():
            # the filehandle already contains the file, so save it
            # directly
            filedata = filehandle.local_filedata()

        _ObjectStore.set_object(bucket=file_bucket,
                                key=file_key,
                                data=filedata)

        if filedata is None:
            # the file is too large to include in the filehandle so
            # we need to use a PAR to upload
            par = _ObjectStore.create_par(bucket=file_bucket,
                                          encrypt_key=encrypt_key,
                                          key=file_key,
                                          readable=False,
                                          writeable=True)

            par_key = "%s/%s" % (_par_root, par.uid())

            data = {"par": par.to_data(),
                    "file_key": file_key,
                    "user_guid": user_guid,
                    "objsize": fileinfo.filesize(),
                    "checksum": fileinfo.checksum()}

            _ObjectStore.set_object_from_json(bucket=metadata_bucket,
                                              key=par_key,
                                              data=data)
        else:
            par = None

        # now save the fileinfo to the object store
        fileinfo.save()

        # return the PAR if we need to have a second-stage of upload
        return (fileinfo.get_filemeta(resolved_acl=file_acl), par)

    def download_complete(self, par_uid, authorisation):
        """Call this function to signify that the file associated with
           the PAR with UID 'par_uid' has been downloaded (must have matching
           authorisation)
        """
        from Acquire.Client import PAR as _PAR
        from Acquire.Client import Authorisation as _Authorisation

        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The authorisation must be of type Authorisation")

        authorisation.verify("downloaded %s" % par_uid)

        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        par_key = "%s/%s" % (_par_root, par_uid)
        file_bucket = self._get_file_bucket()
        metadata_bucket = self._get_metadata_bucket()

        data = _ObjectStore.get_object_from_json(metadata_bucket, par_key)

        par = _PAR.from_data(data["par"])
        file_key = data["file_key"]
        user_guid = data["user_guid"]

        # the same user who requested the PAR must signify that the upload
        # has completed
        if authorisation.user_guid() != user_guid:
            raise PermissionError(
                "Only the user with GUID %s can signify that the "
                "upload is complete. Not user with GUID %s" %
                (user_guid, authorisation.user_guid()))

        if par.is_writeable():
            expected_objsize = data["filesize"]
            expected_checksum = data["checksum"]

            # check that use of the PAR has not changed the file...
            (objsize, checksum) = _ObjectStore.get_size_and_checksum(
                                            file_bucket, file_key)

            if expected_objsize != objsize or expected_checksum != checksum:
                from Acquire.Storage import FileValidationError
                raise FileValidationError(
                    "The file downloaded does not match what was promised. "
                    "size: %s versus %s, checksum: %s versus %s. This "
                    "suggests that the PAR has been used incorrectly!" %
                    (expected_objsize, objsize,
                     expected_checksum, checksum))

                # probably should delete the broken object here...

        _ObjectStore.delete_par(bucket=file_bucket, par=par)
        _ObjectStore.delete_object(bucket=metadata_bucket, key=par_key)

    def download(self, filename, authorisation,
                 version=None, encrypt_key=None):
        """Download the file called filename. This will return a
           FileHandle that describes the file. If the file is
           sufficiently small, then the filedata will be embedded
           into this handle. Otherwise a PAR will be generated and
           also returned to allow the file to be downloaded
           separately. The PAR will be encrypted with 'encrypt_key'
        """
        from Acquire.Client import FileHandle as _FileHandle
        from Acquire.Storage import FileInfo as _FileInfo
        from Acquire.Identity import Authorisation as _Authorisation
        from Acquire.Crypto import PublicKey as _PublicKey
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import string_to_encoded \
            as _string_to_encoded

        if not isinstance(authorisation, _Authorisation):
            raise TypeError("The authorisation must be of type Authorisation")

        if not isinstance(encrypt_key, _PublicKey):
            raise TypeError("The encryption key must be of type PublicKey")

        authorisation.verify("download %s %s" % (self._drive_uid,
                                                 filename))

        user_guid = authorisation.user_guid()

        drive_acl = self.aclrules().resolve(user_guid=user_guid)

        # even if the drive_acl is not readable by this user, they
        # may have read permission for the file...

        # now get the FileInfo for this FileHandle
        fileinfo = _FileInfo.load(drive=self,
                                  filename=filename, version=version)

        # resolve the ACL for the file from this FileHandle
        file_acl = fileinfo.aclrules().resolve(upstream=drive_acl,
                                               user_guid=user_guid)

        if not file_acl.is_readable():
            raise PermissionError(
                "You do not have read permissions for the file. Your file "
                "permissions are %s" % str(file_acl))

        file_bucket = self._get_file_bucket()
        metadata_bucket = self._get_metadata_bucket()

        file_key = fileinfo.latest_version()._file_key()
        filedata = None
        par = None

        if fileinfo.filesize() < 1048576:
            # one-trip download of files that are less than 1 MB
            filedata = _ObjectStore.get_object(file_bucket, file_key)
        else:
            # the file is too large to include in the download so
            # we need to use a PAR to download
            par = _ObjectStore.create_par(bucket=file_bucket,
                                          encrypt_key=encrypt_key,
                                          key=file_key,
                                          readable=True,
                                          writeable=False)

            par_key = "%s/%s" % (_par_root, par.uid())

            data = {"par": par.to_data(),
                    "file_key": file_key,
                    "user_guid": user_guid,
                    "objsize": fileinfo.filesize(),
                    "checksum": fileinfo.checksum()}

            _ObjectStore.set_object_from_json(bucket=metadata_bucket,
                                              key=par_key,
                                              data=data)

        # return the filemeta, and either the filedata or par
        return (fileinfo.get_filemeta(resolved_acl=file_acl), filedata, par)

    def is_opened_by_owner(self):
        """Return whether or not this drive was opened and authorised
           by one of the drive owners
        """
        if self._user_guid is None or (not self._is_authorised):
            return False

        try:
            return self._acls[self._user_guid].is_owner()
        except:
            return False

    def aclrules(self):
        """Return the acl rules for this drive"""
        try:
            return self._aclrules
        except:
            from Acquire.Storage import ACLRules as _ACLRules
            return _ACLRules()

    def set_permission(self, user_guid, aclrule):
        """Set the permission for the user with the passed user_guid
           to 'aclrule". Note that you can only do this if you are the
           owner and this drive was opened in an authorised way. Also
           note that you cannot remove your own ownership permission
           if this would leave the drive without any owners
        """
        if self.is_null():
            return

        from Acquire.Storage import create_aclrules as _create_aclrules
        aclrules = _create_aclrules(aclrule=aclrule, user_guid=user_guid)

        # make sure we have the latest version
        self.load()

        if not self.is_opened_by_owner():
            raise PermissionError(
                "You cannot change user permissions as you are either "
                "not the owner of this drive or you failed to provide "
                "authorisation when you opened the drive")

        # this will append the new rules, ensuring that the change
        # does not leave the drive ownerless
        self._aclrules.append(aclrules, ensure_owner=True)

        self.save()
        self.load()

    def list_files(self, authorisation=None, include_metadata=False):
        """Return the list of FileMeta data for the files contained
           in this Drive. The passed authorisation is needed in case
           the list contents of this drive is not public
        """
        user_guid = None

        if authorisation is not None:
            from Acquire.Client import Authorisation as _Authorisation
            if not isinstance(authorisation, _Authorisation):
                raise TypeError(
                    "The authorisation must be of type Authorisation")

            authorisation.verify("list_files")

            user_guid = authorisation.user_guid()

        drive_acl = self.aclrules().resolve(user_guid=user_guid)

        if not drive_acl.is_readable():
            raise PermissionError(
                "You don't have permission to read this Drive")

        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import encoded_to_string as _encoded_to_string
        from Acquire.Storage import FileMeta as _FileMeta

        metadata_bucket = self._get_metadata_bucket()

        names = _ObjectStore.get_all_object_names(
                    metadata_bucket, "%s/%s" % (_fileinfo_root,
                                                self._drive_uid))

        files = []

        if include_metadata:
            # we need to load all of the metadata info for this file to
            # return to the user
            from Acquire.Storage import FileInfo as _FileInfo

            for name in names:
                data = _ObjectStore.get_object_from_json(metadata_bucket,
                                                         name)
                fileinfo = _FileInfo.from_data(data)
                filemeta = fileinfo.get_filemeta()
                file_acl = filemeta.resolve_acl(upstream=drive_acl,
                                                user_guid=user_guid)

                if file_acl.is_readable() or file_acl.is_writeable():
                    files.append(filemeta)
        else:
            for name in names:
                filename = _encoded_to_string(name.split("/")[-1])
                files.append(_FileMeta(filename=filename))

        return files

    def list_versions(self, filename, authorisation=None,
                      include_metadata=False, **kwargs):
        """Return the list of versions of the file with specified
           filename. If 'include_metadata' is true then this will
           load full metadata for each version. This will return
           a sorted list of FileMeta objects. The passed authorisation
           is needed in case the version info is not public
        """
        user_guid = None

        if authorisation is not None:
            from Acquire.Client import Authorisation as _Authorisation
            if not isinstance(authorisation, _Authorisation):
                raise TypeError(
                    "The authorisation must be of type Authorisation")

            authorisation.verify("list_versions %s" % filename)

            user_guid = authorisation.user_guid()

        drive_acl = self.aclrules().resolve(user_guid=user_guid, **kwargs)

        if not drive_acl.is_readable():
            raise PermissionError(
                "You don't have permission to read this Drive")

        from Acquire.Storage import FileInfo as _FileInfo
        versions = _FileInfo.list_versions(drive=self,
                                           filename=filename,
                                           include_metadata=include_metadata)

        result = []

        for version in versions:
            if version.aclrules() is not None:
                acl = version.resolve_acl(upstream=drive_acl,
                                          user_guid=user_guid,
                                          **kwargs)

                if acl.is_readable() or acl.is_writeable():
                    result.append(version)
            else:
                result.append(version)

        # return the versions sorted in upload order
        versions.sort(key=lambda x: x.uploaded_when())

        return versions

    def load(self):
        """Load the metadata about this drive from the object store"""
        if self.is_null():
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        bucket = _get_service_account_bucket()

        drive_key = self._drive_key()

        try:
            data = _ObjectStore.get_object_from_json(bucket, drive_key)
        except:
            data = None

        if data is None:
            if self._user_guid is None:
                # we cannot create the drive as we don't know who
                # requested it
                raise PermissionError(
                    "Cannot create the DriveInfo for a new drive as the "
                    "original request did not specify the creating user")
            elif not self._is_authorised:
                # we cannot create the drive if the request was not
                # authorised
                raise PermissionError(
                    "Cannot create the DriveInfo for a new drive as the "
                    "original request was not authorised by the user")

            # create a new drive and save it...
            from Acquire.Storage import ACLRule as _ACLRule
            from Acquire.Storage import create_aclrules as _create_aclrules

            # by default this user is the drive's owner
            self._aclrules = _create_aclrules(user_guid=self._user_guid,
                                              aclrule=_ACLRule.owner())

            data = self.to_data()

            data = _ObjectStore.set_ins_object_from_json(bucket, drive_key,
                                                         data)

        from copy import copy as _copy
        other = DriveInfo.from_data(data)

        user_guid = self._user_guid
        is_authorised = self._is_authorised

        self.__dict__ = _copy(other.__dict__)

        self._user_guid = user_guid
        self._is_authorised = is_authorised

    def save(self):
        """Save the metadata about this drive to the object store"""
        if self.is_null():
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        bucket = _get_service_account_bucket()

        drive_key = self._drive_key()

        data = self.to_data()
        _ObjectStore.set_object_from_json(bucket, drive_key, data)

    def to_data(self):
        """Return a json-serialisable dictionary for this object"""
        data = {}

        if not self.is_null():
            from Acquire.ObjectStore import dict_to_string as _dict_to_string
            data["uid"] = self._drive_uid

            if self._aclrules is not None:
                data["aclrules"] = self._aclrules.to_data()

        return data

    @staticmethod
    def from_data(data):
        """Return a DriveInfo constructed from the passed json-deserialised
           dictionary
        """
        info = DriveInfo()

        if data is None or len(data) == 0:
            return info

        info._drive_uid = str(data["uid"])

        if "aclrules" in data:
            from Acquire.Storage import ACLRules as _ACLRules
            info._aclrules = _ACLRules.from_data(data["aclrules"])

        return info




__all__ = ["lazy_import"]


class lazy_import:
    """This is not lazy_import, but instead a thin stub that matches the
       API but DOES NOT lazy_import anything. This imports at call time.
       Use this module if you are running a python installation that
       does not have lazy_import installed, e.g. because you don't want
       to install any GPL modules
    """
    @staticmethod
    def lazy_module(m):
        return __import__(m, fromlist=[''])

    @staticmethod
    def lazy_function(f):
        module_name, unit_name = f.rsplit('.', 1)
        return getattr(__import__(module_name, fromlist=['']), unit_name)

    @staticmethod
    def lazy_class(c):
        return lazy_import.lazy_function(c)




from Acquire.Stubs import lazy_import as _lazy_import

import os as _os
import sys as _sys
import json as _json

_getpass = _lazy_import.lazy_module("getpass")
_glob = _lazy_import.lazy_module("glob")
_re = _lazy_import.lazy_module("re")

# use a variable so we can monkey-patch while testing
_input = input

# whether or not we are in testing mode
_is_testing = False

__all__ = ["Wallet"]


def _read_json(filename):
    """Return a json-decoded dictionary from the data written
       to 'filename'
    """
    global _is_testing
    if _is_testing:
        return {}

    with open(filename, "rb") as FILE:
        s = FILE.read().decode("utf-8")
        return _json.loads(s)


def _write_json(data, filename):
    """Write the passed json-encodable dictionary to 'filename'"""
    global _is_testing
    if _is_testing:
        return

    s = _json.dumps(data)
    with open(filename, "wb") as FILE:
        FILE.write(s.encode("utf-8"))


def _read_service(filename):
    """Read and return the service written to 'filename'"""
    from Acquire.Client import Service as _Service
    return _Service.from_data(_read_json(filename))


def _write_service(service, filename):
    """Write the passed service to 'filename'"""
    _write_json(service.to_data(), filename)


class Wallet:
    """This class holds a wallet that can be used to simplify
       sending passwords and one-time-password (OTP) codes
       to an acquire identity service.

       This holds a wallet of passwords and (optionally)
       OTP secrets that are encrypted using a local keypair
       that is unlocked by a password supplied by the user locally.
    """
    def __init__(self):
        self._wallet_key = None
        self._get_wallet_key()
        self._cache = {}
        self._service_info = {}
        self._manual_password = False
        self._manual_otpcode = False

    @staticmethod
    def _wallet_dir():
        """Directory containing all of the wallet files"""
        homedir = _os.path.expanduser("~")

        walletdir = "%s/.acquire" % homedir

        if not _os.path.exists(walletdir):
            _os.mkdir(walletdir)

        return walletdir

    def _create_wallet_key(self, filename):
        """Create a new wallet key for the user"""

        password = _getpass.getpass(
                     prompt="Please enter a password to encrypt your wallet: ")

        from Acquire.Client import PrivateKey as _PrivateKey
        key = _PrivateKey()

        bytes = key.bytes(password)

        password2 = _getpass.getpass(prompt="Please confirm the password: ")

        if password != password2:
            print("The passwords don't match. Please try again.")
            self._create_wallet_key(filename)
            return

        # the passwords match - write this to the file
        with open(filename, "wb") as FILE:
            FILE.write(bytes)

        return key

    def _get_wallet_key(self):
        """Return the private key used to encrypt everything in the wallet.
           This will ask for the users password
        """
        if self._wallet_key:
            return self._wallet_key

        wallet_dir = Wallet._wallet_dir()

        keyfile = "%s/wallet_key.pem" % wallet_dir

        if not _os.path.exists(keyfile):
            self._wallet_key = self._create_wallet_key(keyfile)
            return self._wallet_key

        # read the keyfile and decrypt
        with open(keyfile, "rb") as FILE:
            bytes = FILE.read()

        # get the user password
        wallet_key = None
        while not wallet_key:
            password = _getpass.getpass(
                            prompt="Please enter your wallet password: ")

            try:
                from Acquire.Client import PrivateKey as _PrivateKey
                wallet_key = _PrivateKey.read_bytes(bytes, password)
            except:
                print("Invalid password. Please try again.")

        self._wallet_key = wallet_key
        return wallet_key

    @staticmethod
    def _get_userfile(username, identity_url):
        """Return the userfile for the passed username logging into the
           passed identity url"""
        from Acquire.ObjectStore import string_to_safestring \
            as _string_to_safestring
        return "%s/user_%s_%s_encrypted" % (
            Wallet._wallet_dir(),
            _string_to_safestring(username),
            _string_to_safestring(identity_url))

    @staticmethod
    def remove_user_info(username):
        """Call this function to remove the userinfo associated
           with the account 'username' for all identity services
        """
        wallet_dir = Wallet._wallet_dir()

        userfiles = _glob.glob("%s/user_*_encrypted" % wallet_dir)

        for userfile in userfiles:
            try:
                userinfo = self._read_userfile(userfile)
                if userinfo["username"] == username:
                    _os.unlink(userfile)
            except:
                pass

    def _read_userfile(self, filename):
        """Read all info from the passed userfile"""
        try:
            return self._cache[filename]
        except:
            pass

        if not _os.path.exists(filename):
            return

        data = _read_json(filename)

        from Acquire.ObjectStore import string_to_bytes \
            as _string_to_bytes

        try:
            data["password"] = _string_to_bytes(data["password"])
        except:
            pass

        try:
            data["otpsecret"] = _string_to_bytes(data["otpsecret"])
        except:
            pass

        self._cache[filename] = data
        return data

    def _read_userinfo(self, username, identity_url):
        """Read all info for the passed user at the identity service
           reached at 'identity_url'"""
        return self._read_userfile(Wallet._get_userfile(username,
                                                        identity_url))

    def _get_username(self):
        """Function to find a username automatically, of if that fails,
           to ask the user
        """
        wallet_dir = Wallet._wallet_dir()

        userfiles = _glob.glob("%s/user_*_encrypted" % wallet_dir)

        userinfos = {}

        from Acquire.ObjectStore import safestring_to_string \
            as _safestring_to_string

        for userfile in userfiles:
            try:
                userinfo = self._read_userfile(userfile)
                username = _safestring_to_string(userinfo["username"])
                userinfos[username] = userinfo
            except:
                pass

        usernames = list(userinfos.keys())
        usernames.sort()

        if len(usernames) == 1:
            return usernames[0]

        if len(usernames) == 0:
            return _input("Please type your username: ")

        print("Please choose the account by typing in its number, "
              "or type a new username if you want a different account.")

        for (i, username) in enumerate(usernames):
            print("[%d] %s" % (i+1, username))

        username = None

        while not username:
            print("Make your selection: ", end="")

            reply = _input(
                    "Make your selection (1 to %d) " %
                    (len(usernames))
                )

            try:
                idx = int(reply) - 1

                if idx < 0 or idx >= len(usernames):
                    print("Invalid account. Try again...")
                else:
                    username = usernames[idx]
            except:
                username = None

        return username

    def _get_user_password(self, username, identity_url):
        """Get the user password for the passed user on the passed
           identity_url. If remember_device then save the
           password in the wallet if it is not already there
        """
        try:
            userinfo = self._read_userinfo(username, identity_url)

            if userinfo:
                password = userinfo["password"]
                self._manual_password = False

                # this needs to be decrypted
                return self._wallet_key.decrypt(password)
        except:
            pass

        self._manual_password = True
        return _getpass.getpass(prompt="Please enter the login password: ")

    def _get_otpcode(self, username, identity_url):
        """Get the OTP code for the passed user on the passed identity
           service url"""
        try:
            userinfo = self._read_userinfo(username, identity_url)

            if userinfo:
                from Acquire.Client import OTP as _OTP
                secret = self._wallet_key.decrypt(userinfo["otpsecret"])
                device_uid = userinfo["device_uid"]
                self._manual_otpcode = False
                self._device_uid = device_uid
                return _OTP(secret=secret).generate()
        except:
            pass

        self._manual_otpcode = True
        self._device_uid = None
        return _getpass.getpass(
                    prompt="Please enter the one-time-password code: ")

    @staticmethod
    def add_service(service):
        """Add a cached service info for the passed service. If it
           already exists, then this verifies that the added service
           is the same as the previously-seen service
        """
        global _is_testing
        if _is_testing:
            return service

        from Acquire.ObjectStore import string_to_safestring \
            as _string_to_safestring

        service_file = "%s/service_%s" % (
            Wallet._wallet_dir(),
            _string_to_safestring(service.canonical_url()))

        existing_service = None

        try:
            existing_service = _read_service(service_file)
        except:
            pass

        if existing_service is not None:
            if service.validation_string() == \
               existing_service.validation_string():
                return service
            elif service.is_evolution_of(existing_service):
                # the service has evolved - this is ok
                _write_service(service, service_file)
                return service
            else:
                reply = _input(
                    "This is a service you have seen before, but "
                    "it has changed?\n\n"
                    "URL = %s (%s)\n"
                    "UID = %s (%s)\n"
                    "public_key fingerprint = %s (%s)\n"
                    "public_certificate fingerprint = %s (%s)\n\n"
                    "verification string = %s (%s)\n\n"
                    "\nDo you trust this updated service? y/n " %
                    (service.canonical_url(),
                     existing_service.canonical_url(),
                     service.uid(), existing_service.uid(),
                     service.public_key().fingerprint(),
                     existing_service.public_key().fingerprint(),
                     service.public_certificate().fingerprint(),
                     existing_service.public_certificate().fingerprint(),
                     service.validation_string(),
                     existing_service.validation_string())
                )

                if reply[0].lower() == 'y':
                    print("Now trusting %s" % str(service))
                else:
                    print("Not trusting this service!")
                    raise PermissionError(
                        "We do not trust the service '%s'" % str(service))

                # We trust the service, so save this for future reference
                _write_service(service, service_file)
                return service

        reply = _input(
                    "This is a new service that you have not seen before.\n\n"
                    "URL = %s\n"
                    "UID = %s\n"
                    "public_key fingerprint = %s\n"
                    "public_certificate fingerprint = %s\n\n"
                    "verification string = %s\n\n"
                    "\nDo you trust this service? y/n " %
                    (service.canonical_url(),
                     service.uid(),
                     service.public_key().fingerprint(),
                     service.public_certificate().fingerprint(),
                     service.validation_string())
                )

        if reply[0].lower() == 'y':
            print("Now trusting %s" % str(service))
        else:
            print("Not trusting this service!")
            raise PermissionError(
                "We do not trust the service '%s'" % str(service))

        # We trust the service, so save this for future reference
        _write_service(service, service_file)

        return service

    @staticmethod
    def get_services():
        """Return all of the trusted services known to this wallet"""
        global _is_testing
        if _is_testing:
            return []

        service_files = _glob.glob("%s/service_*" % Wallet._wallet_dir())

        services = []

        for service_file in service_files:
            services.append(_read_service(service_file))

        return services

    @staticmethod
    def get_service(service_url):
        """Return the service at 'service_url'. This will return the
           cached service if it exists, or will add a new service if
           the user so wishes
        """
        global _is_testing
        if _is_testing:
            from Acquire.Service import get_remote_service as \
                _get_remote_service
            return _get_remote_service(service_url)

        from Acquire.ObjectStore import string_to_safestring \
            as _string_to_safestring

        service_file = "%s/service_%s" % (
            Wallet._wallet_dir(),
            _string_to_safestring(service_url))

        existing_service = None

        try:
            existing_service = _read_service(service_file)
        except:
            pass

        if existing_service is not None:
            # check if the keys need rotating - if they do, load up
            # the new keys and save them to the service file...
            if existing_service.should_refresh_keys():
                existing_service.refresh_keys()
                _write_service(existing_service, service_file)

            return existing_service
        else:
            from Acquire.Service import get_remote_service as \
                _get_remote_service

            service = _get_remote_service(service_url)
            return Wallet.add_service(service)

    @staticmethod
    def remove_all_services():
        """Remove all trusted services from this Wallet"""
        global _is_testing
        if _is_testing:
            return

        service_files = _glob.glob("%s/service_*" % Wallet._wallet_dir())

        for service_file in service_files:
            if _os.path.exists(service_file):
                _os.unlink(service_file)

        # clear cache to force a new lookup
        from ._service import _cache_service_lookup
        _cache_service_lookup.clear()

    @staticmethod
    def remove_service(service):
        """Remove the cached service info for the passed service"""
        global _is_testing
        if _is_testing:
            return

        if isinstance(service, str):
            service_url = service
        else:
            service_url = service.canonical_url()

        from Acquire.ObjectStore import string_to_safestring \
            as _string_to_safestring

        service_file = "%s/service_%s" % (
            Wallet._wallet_dir(),
            _string_to_safestring(service_url))

        if _os.path.exists(service_file):
            _os.unlink(service_file)

        # clear cache to force a new lookup
        from ._service import _cache_service_lookup
        _cache_service_lookup.clear()

    def send_password(self, url, username=None, remember_password=True,
                      remember_device=None, dryrun=None):
        """Send a password and one-time code to the supplied login url"""

        self._manual_password = False
        self._manual_otpcode = False

        if not remember_password:
            remember_device = False

        # the login URL is of the form "server/code"
        words = url.split("/")
        identity_service = "/".join(words[0:-1])
        short_uid = words[-1].split("=")[-1]

        # now get the service
        service = Wallet.get_service(identity_service)

        if not service.can_identify_users():
            from Acquire.Client import LoginError
            raise LoginError(
                "Service '%s' is unable to identify users! "
                "You cannot log into something that is not "
                "a valid identity service!" % (service))

        if not username:
            # choose a username from any existing files...
            username = self._get_username()

        print("Logging in using username '%s'" % username)
        password = self._get_user_password(username, service.canonical_url())
        otpcode = self._get_otpcode(username, service.canonical_url())

        print("\nLogging in to '%s', session '%s'..." % (
              service.canonical_url(), short_uid), end="")
        _sys.stdout.flush()

        if dryrun:
            print("Calling %s with username=%s, password=%s, otpcode=%s, "
                  "remember_device=%s, device_uid=%s, short_uid=%s" %
                  (service.canonical_url(), username, password, otpcode,
                   remember_device, self._device_uid, short_uid))
            return

        try:
            function = "login"
            args = {"username": username,
                    "password": password,
                    "otpcode": otpcode,
                    "remember_device": remember_device,
                    "device_uid": self._device_uid,
                    "short_uid": short_uid}

            response = service.call_function(function=function, args=args)
            print("SUCCEEDED!")
            _sys.stdout.flush()
        except Exception as e:
            print("FAILED!")
            _sys.stdout.flush()
            from Acquire.Client import LoginError
            raise LoginError("Failed to log in. %s" % e.args)

        if remember_password:
            try:
                provisioning_uri = response["provisioning_uri"]
            except:
                provisioning_uri = None

            try:
                device_uid = response["device_uid"]
            except:
                device_uid = None

            otpsecret = None

            if provisioning_uri:
                try:
                    otpsecret = _re.search(r"secret=([\w\d+]+)&issuer",
                                           provisioning_uri).groups()[0]
                except:
                    pass

            try:
                user_info = self._read_userinfo(username,
                                                service.canonical_url())
            except:
                user_info = {}

            if user_info is None:
                user_info = {}

            pubkey = self._wallet_key.public_key()

            must_write = self._manual_password

            if otpsecret:
                if self._manual_otpcode:
                    must_write = True

            if must_write:
                from Acquire.ObjectStore import string_to_safestring \
                    as _string_to_safestring
                from Acquire.ObjectStore import bytes_to_string \
                    as _bytes_to_string

                user_info["username"] = _string_to_safestring(username)
                user_info["password"] = _bytes_to_string(
                                              pubkey.encrypt(
                                                  password.encode("utf-8")))

                if otpsecret:
                    user_info["otpsecret"] = _bytes_to_string(
                                                pubkey.encrypt(
                                                   otpsecret.encode("utf-8")))
                    user_info["device_uid"] = device_uid

                _write_json(data=user_info, filename=Wallet._get_userfile(
                            username, service.canonical_url()))

        self._manual_password = False
        self._manual_otpcode = False

        return response




import uuid as _uuid
from copy import copy as _copy
import os as _os

from Acquire.Service import Service as _Service

from ._errors import StorageServiceError

__all__ = ["StorageService"]


class StorageService(_Service):
    """This is a specialisation of Service for Storage Services"""
    def __init__(self, other=None):
        if isinstance(other, _Service):
            self.__dict__ = _copy(other.__dict__)

            if not self.is_storage_service():
                from Acquire.Storage import StorageServiceError
                raise StorageServiceError(
                    "Cannot construct an StorageService from "
                    "a service which is not an storage service!")

            # the storage service must define the ID for the compartment
            # in which user data will be stored
            self._storage_compartment_id = _os.getenv("STORAGE_COMPARTMENT")
        else:
            _Service.__init__(self)
            self._storage_compartment_id = None

    def storage_compartment(self):
        """Return the ID of the compartment in which user data will be
           stored. This should be a different compartment to the one used
           to store management data for the storage service"""
        if self._storage_compartment_id is None:
            from Acquire.Storage import StorageServiceError
            raise StorageServiceError(
                "The ID of the compartment for the storage account has not "
                "been set. This should have been set when the StorageService "
                "was constructed, e.g. via the STORAGE_COMPARTMENT env "
                "variable configured via Fn config")

        try:
            return self._storage_compartment_id
        except:
            pass

    def _call_local_function(self, function, args):
        """Internal function called to short-cut local 'remote'
           function calls
        """
        from storage.route import storage_functions as _storage_functions
        from admin.handler import create_handler as _create_handler
        handler = _create_handler(_storage_functions)
        return handler(function, args)




__all__ = ["JobSheet"]


class JobSheet:
    """This class holds a complete record of a job that the access
       service has been asked to perform.
    """
    def __init__(self, job=None, authorisation=None):
        if job is not None:
            from Acquire.Identity import Authorisation as _Authorisation
            if not isinstance(authorisation, _Authorisation):
                raise TypeError("You can only authorise a reqeust with "
                                "a valid Authorisation object")
            authorisation.verify(job.fingerprint())
            from Acquire.ObjectStore import create_uuid as _create_uuid
            self._job = job
            self._authorisation = authorisation
            self._uid = _create_uuid()
        else:
            self._uid = None

    def is_null(self):
        """Return whether or not this JobSheet is null"""
        return self._uid is None

    def uid(self):
        """Return the UID of this JobSheet"""
        return self._uid

    def total_cost(self):
        """Return the total maximum quoted cost for this job. The
           total cost to run the job must not exceed this
        """
        if self.is_null():
            return 0
        else:
            return 10

    def job(self):
        """Return the original job request"""
        if self.is_null():
            return None
        else:
            return self._job

    def authorisation(self):
        """Return the original authorisation for this job"""
        if self.is_null():
            return None
        else:
            return self._authorisation

    def set_payment(self, cheque):
        """Pass in and cash that contains the source of value of
           paying for this job. This will cash the cheque and store
           the value within credit notes held in this job sheet
        """
        if self.is_null():
            from Acquire.Accounting import PaymentError
            raise PaymentError("You cannot try to pay for a null job!")

        from Acquire.Client import Cheque as _Cheque
        if not isinstance(cheque, _Cheque):
            raise TypeError("You must pass a valid Cheque as payment "
                            "for a job")

        try:
            credit_notes = cheque.cash(spend=self.total_cost(),
                                       resource=self.job().fingerprint())
        except Exception as e:
            from Acquire.Service import exception_to_string
            from Acquire.Accounting import PaymentError
            raise PaymentError(
                "Problem cashing the cheque used to pay for the calculation: "
                "\n\nCAUSE: %s" % exception_to_string(e))

        if credit_notes is None or len(credit_notes) == 0:
            from Acquire.Accounting import PaymentError
            raise PaymentError("Cannot be paid!")

        # save these credit_notes so that they are not lost
        self._credit_notes = credit_notes

        # save the the object store so we always have a record of this value
        self.save()

    def request_services(self):
        """Request all of the services needed to perform the job. This
           will contract a compute service and a storage service to run
           the job.

           The storage service will provide (1) a file upload PAR to enable
           the user to upload the input, and (2) a bucket write PAR to
           enable the compute service to write the output

           The compute service will be supplied with the bucket write PAR
           from the storage service and will supply a run calculation PAR
           to enable the user to trigger the start of the job

           This returns the upload PAR the user must use to upload the
           input, the run PAR that the user must call after uploading
           input to trigger the start of the calculation, and the expiry
           date when both of these PARs will become invalid (i.e. the user
           must trigger both of them before that date)
        """
        # make the requests, make the payments

        # save so we don't lost the debit notes or any value
        self.save()

        from Acquire.Client import PAR as _PAR
        from Acquire.ObjectStore import get_datetime_future \
            as _get_datetime_future

        return (_PAR(), _PAR(), _get_datetime_future(hours=1))

    def save(self):
        """Save this JobSheet to the object store"""
        from Acquire.Service import assert_running_service \
            as _assert_running_service

        _assert_running_service()

        if self.is_null():
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        key = "jobsheets/%s" % self.uid()
        _ObjectStore.set_object_from_json(bucket, key, self.to_data())

    @staticmethod
    def load(uid):
        """Return the JobSheet with specified uid loaded from the
           ObjectStore
        """
        from Acquire.Service import assert_running_service \
            as _assert_running_service

        _assert_running_service()

        if uid is None:
            return

        from Acquire.Service import get_service_account_bucket \
            as _get_service_account_bucket

        bucket = _get_service_account_bucket()

        from Acquire.ObjectStore import ObjectStore as _ObjectStore

        key = "jobsheets/%s" % str(uid)
        data = _ObjectStore.get_object_from_json(bucket, key)
        return JobSheet.from_data(data)

    def to_data(self):
        """Return this converted to a json-serialisable dictionary"""
        data = {}

        if self.is_null():
            return data

        from Acquire.ObjectStore import list_to_string as _list_to_string

        data["uid"] = self.uid()
        data["job"] = self.job().to_data()
        data["authorisation"] = self.authorisation().to_data()
        data["credit_notes"] = _list_to_string(self._credit_notes)

        return data

    @staticmethod
    def from_data(data):
        """Return a JobSheet constructed from the passed json-deserialised
           dictionary
        """
        j = JobSheet()

        if (data and len(data) > 0):
            from Acquire.Access import RunRequest as _RunRequest
            from Acquire.Client import Authorisation as _Authorisation
            from Acquire.Accounting import CreditNote as _CreditNote
            from Acquire.ObjectStore import string_to_list \
                as _string_to_list

            j._uid = str(data["uid"])
            j._job = _RunRequest.from_data(data["job"])
            j._authorisation = _Authorisation.from_data(
                                            data["authorisation"])
            j._credit_notes = _string_to_list(data["credit_notes"],
                                              _CreditNote)

        return j




__all__ = ["LineItem"]


class LineItem:
    """This class holds the data for a line item in the account. This holds
       basic information about the item, e.g. its UID and authorisation
    """
    def __init__(self, uid=None, authorisation=None):
        self._uid = uid

        if authorisation is not None:
            from Acquire.Identity import Authorisation as _Authorisation

            if not isinstance(authorisation, _Authorisation):
                raise TypeError("Authorisation must be of type Authorisation!")

        self._authorisation = authorisation

    def __str__(self):
        return "LineItem(uid=%s)" % self._uid

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._uid == other._uid
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def is_null(self):
        """Return whether or not this is a null line item"""
        return self._uid is None

    def uid(self):
        """Return the UID of the TransactionRecord that provides
           more information about this line item in the ledger
        """
        return self._uid

    def authorisation(self):
        """Return the authorisation that was used to authorise this action"""
        return self._authorisation

    def to_data(self):
        """Return this object as a dictionary that can be serialised to json"""
        data = {}

        if not self.is_null():
            data["uid"] = self._uid

            if self._authorisation is not None:
                data["authorisation"] = self._authorisation.to_data()

        return data

    @staticmethod
    def from_data(data):
        """Return a LineItem constructed from the json-decoded dictionary"""
        l = LineItem()

        if (data and len(data) > 0):
            l._uid = data["uid"]

            if "authorisation" in data:
                from Acquire.Identity import Authorisation as _Authorisation
                l._authorisation = _Authorisation.from_data(
                    data["authorisation"])
            else:
                l._authorisation = None

        return l




import base64 as _base64

from cachetools import cached as _cached
from cachetools import LRUCache as _LRUCache

_cache_local_serviceinfo = _LRUCache(maxsize=5)
_cache_local_serviceinfos = _LRUCache(maxsize=5)
_cache_remote_serviceinfo = _LRUCache(maxsize=20)

__all__ = ["get_trusted_service", "get_trusted_services",
           "get_remote_service", "get_checked_remote_service",
           "clear_services_cache"]


def clear_services_cache():
    """Clear the caches of loaded services"""
    _cache_local_serviceinfo.clear()
    _cache_remote_serviceinfo.clear()


# Cached as the list of all trusted remote service information
# will not change too often
@_cached(_cache_local_serviceinfos)
def get_trusted_services():
    """Return a dictionary of all trusted services indexed by
       their type
    """
    from Acquire.Service import is_running_service as _is_running_service

    if _is_running_service():
        from Acquire.Service import get_this_service as _get_this_service
        from Acquire.Service import Service as _Service

        from Acquire.Service import get_service_account_bucket as \
            _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import url_to_encoded as \
            _url_to_encoded

        # we already trust ourselves
        service = _get_this_service()

        trusted_services = {}
        trusted_services[service.service_type()] = [service]

        bucket = _get_service_account_bucket()

        uidkey = "_trusted/uid/"
        datas = _ObjectStore.get_all_objects(bucket, uidkey)

        for data in datas:
            remote_service = _Service.from_data(data)

            if remote_service.should_refresh_keys():
                # need to update the keys in our copy of the service
                remote_service.refresh_keys()
                key = "%s/%s" % (uidkey, remote_service.uid())
                _ObjectStore.set_object_from_json(bucket, key,
                                                  remote_service.to_data())

            if remote_service.service_type() in datas:
                datas[remote_service.service_type()].append(remote_service)
            else:
                datas[remote_service.service_type()] = [remote_service]

        return datas
    else:
        # this is running on the client
        from Acquire.Client import Wallet as _Wallet
        return _Wallet.get_services()


# Cached as the remote service information will not change too often
@_cached(_cache_local_serviceinfo)
def get_trusted_service(service_url=None, service_uid=None):
    """Return the trusted service info for the service with specified
       service_url or service_uid"""
    from Acquire.Service import is_running_service as _is_running_service

    if _is_running_service():
        from Acquire.Service import get_this_service as _get_this_service
        from Acquire.Service import Service as _Service
        from Acquire.Service import get_service_account_bucket as \
            _get_service_account_bucket
        from Acquire.ObjectStore import ObjectStore as _ObjectStore
        from Acquire.ObjectStore import url_to_encoded as \
            _url_to_encoded

        service = _get_this_service()

        if service.canonical_url() == service_url:
            # we trust ourselves :-)
            return service

        bucket = _get_service_account_bucket()

        uidkey = None

        if service_uid is not None:
            uidkey = "_trusted/uid/%s" % service_uid
            data = _ObjectStore.get_object_from_json(bucket, uidkey)
        elif service_url is not None:
            urlkey = "_trusted/url/%s" % _url_to_encoded(service_url)
            uidkey = _ObjectStore.get_string_object(bucket, urlkey)
            if uidkey is not None:
                data = _ObjectStore.get_object_from_json(bucket, uidkey)
        else:
            data = None

        if data is None:
            from Acquire.Service import ServiceAccountError
            if service_uid is not None:
                raise ServiceAccountError(
                    "We do not trust the service with UID '%s'" %
                    service_uid)
            else:
                raise ServiceAccountError(
                    "We do not trust the service at URL '%s'" %
                    service_url)

        remote_service = _Service.from_data(data)

        if remote_service.should_refresh_keys():
            # need to update the keys in our copy of the service
            remote_service.refresh_keys()

            if uidkey is not None:
                _ObjectStore.set_object_from_json(bucket, uidkey,
                                                  remote_service.to_data())

        return remote_service
    else:
        # this is running on the client
        from Acquire.Client import Service as _Service
        return _Service(service_url=service_url, service_uid=service_uid)


# Cached to stop us sending too many requests for info to remote services
@_cached(_cache_remote_serviceinfo)
def get_remote_service(service_url):
    """This function returns the service info for the service at
       'service_url'
    """
    from Acquire.Crypto import get_private_key as _get_private_key
    from Acquire.Service import call_function as _call_function
    from Acquire.Service import Service as _Service

    key = _get_private_key("function")

    try:
        response = _call_function(service_url, response_key=key)
    except Exception as e:
        from Acquire.Service import ServiceError
        raise ServiceError("Cannot get information about '%s': %s" %
                           (service_url, str(e)))

    from Acquire.Crypto import SignatureVerificationError \
        as _SignatureVerificationError

    try:
        return _Service.from_data(response["service_info"],
                                  verify_data=True)
    except _SignatureVerificationError:
        raise
    except Exception as e:
        from Acquire.Service import ServiceError
        raise ServiceError(
                "Cannot extract service info for '%s' from '%s': %s" %
                (service_url, str(response), str(e)))


def get_checked_remote_service(service_url, public_cert):
    """This function returns the service info for the service at
       'service_url'. This checks that the info has been signed
       correctly by the passed public certificate
    """
    from Acquire.Crypto import get_private_key as _get_private_key
    from Acquire.Service import call_function as _call_function
    from Acquire.Service import Service as _Service

    key = _get_private_key("function")

    try:
        response = _call_function(service_url, response_key=key,
                                  public_cert=public_cert)
    except Exception as e:
        from Acquire.Service import ServiceError
        raise ServiceError("Cannot get information about '%s': %s" %
                           (service_url, str(e)))

    from Acquire.Crypto import SignatureVerificationError \
        as _SignatureVerificationError

    try:
        return _Service.from_data(response["service_info"],
                                  verify_data=True)
    except _SignatureVerificationError:
        raise
    except Exception as e:
        from Acquire.Service import ServiceError
        raise ServiceError(
                "Cannot extract service info for '%s' from '%s': %s" &
                (service_url, str(response), str(e)))




from Acquire.Service import ServiceError as _ServiceError

__all__ = ["StorageServiceError", "MissingDriveError", "MissingFileError",
           "MissingVersionError", "FileValidationError"]


class StorageServiceError(_ServiceError):
    pass


class MissingDriveError(Exception):
    pass


class MissingFileError(Exception):
    pass


class MissingVersionError(Exception):
    pass


class FileValidationError(Exception):
    pass




import json as _json
import base64 as _base64
import datetime as _datetime
import uuid as _uuid
import os as _os
import sys as _sys

if _sys.version_info.major < 3:
    raise ImportError("Acquire requires Python 3.6 minimum")

if _sys.version_info.minor < 6:
    raise ImportError("Acquire requires Python 3.6 minimum")

if _sys.version_info.major == 3 and _sys.version_info.minor == 6:
    try:
        from backports.datetime_fromisoformat import MonkeyPatch \
            as _MonkeyPatch
        _MonkeyPatch.patch_fromisoformat()
    except:
        raise ImportError(
            "backports-datetime-fromisoformat must be installed "
            "on Python < 3.7. Please run 'pip install "
            "backports-datetime-fromisoformat' or update to a newer "
            "version of Python")

__all__ = ["bytes_to_string", "string_to_bytes",
           "string_to_encoded", "encoded_to_string",
           "url_to_encoded", "encoded_to_url",
           "decimal_to_string", "string_to_decimal",
           "datetime_to_string", "string_to_datetime",
           "date_to_string", "string_to_date",
           "time_to_string", "string_to_time",
           "get_datetime_now", "datetime_to_datetime",
           "string_to_safestring", "safestring_to_string",
           "string_to_list", "list_to_string",
           "string_to_dict", "dict_to_string",
           "string_to_filepath", "string_to_filepath_parts",
           "get_datetime_future",
           "get_datetime_now_to_string",
           "date_and_time_to_datetime",
           "create_uuid"]


def create_uuid():
    """Return a newly created random uuid. This is highly likely
       to be globally unique
    """
    return str(_uuid.uuid4())


def string_to_encoded(s):
    """Return the passed unicode string encoded to a safely
       encoded base64 utf-8 string"""
    return bytes_to_string(s.encode("utf-8"))


def encoded_to_string(b):
    """Return the passed encoded base64 utf-8 string converted
       back into a unicode string"""
    return string_to_bytes(b).decode("utf-8")


def url_to_encoded(url):
    """Return an encoding of the passed url that is safe to use
       as a name, filename or key in an object store
    """
    return _base64.b64encode(url.encode("utf-8")).decode("utf-8")


def encoded_to_url(e):
    """Decode the passed encoded data back to the URL. This will only
       produce valid output for inputs created using url_to_encoded
    """
    return _base64.b64decode(e.encode("utf-8")).decode("utf-8")


def bytes_to_string(b):
    """Return the passed binary bytes safely encoded to
       a base64 utf-8 string"""
    if b is None:
        return None
    else:
        return _base64.b64encode(b).decode("utf-8")


def string_to_bytes(s):
    """Return the passed base64 utf-8 encoded binary data
       back converted from a string back to bytes. Note that
       this can only convert strings that were encoded using
       bytes_to_string - you cannot use this to convert
       arbitrary strings to bytes"""
    if s is None:
        return None
    else:
        return _base64.b64decode(s.encode("utf-8"))


def decimal_to_string(d):
    """Return the passed decimal number encoded as a string that
       can be safely serialised via json
    """
    return str(d)


def string_to_decimal(s):
    """Return the decimal that had been encoded via 'decimal_to_string'.
       This string must have been created via 'decimal_to_string'
    """
    from Acquire.Accounting import create_decimal as _create_decimal
    return _create_decimal(s)


def datetime_to_string(d):
    """Return the passed datetime encoded to a string. This will be a
       standard iso-formatted time in the UTC timezone (converting
       to UTC if the passed datetime is for another timezone)
    """
    if d.tzinfo is None:
        d = d.replace(tzinfo=_datetime.timezone.utc)
    else:
        d = d.astimezone(_datetime.timezone.utc)

    # the datetime is in UTC, so write out the string without
    # the unnecessary +00:00
    return d.replace(tzinfo=None).isoformat()


def datetime_to_datetime(d):
    """Return the passed datetime as a datetime that is clean
       and usable by Acquire. This will move the datetime to UTC,
       adding the timezone if this is missing
    """
    if not isinstance(d, _datetime.datetime):
        raise TypeError(
            "The passed object '%s' is not a valid datetime" % str(d))

    if d.tzinfo is None:
        return d.replace(tzinfo=_datetime.timezone.utc)
    else:
        return d.astimezone(_datetime.timezone.utc)


def date_and_time_to_datetime(date, time=_datetime.time(0)):
    """Return the passed date and time as a UTC datetime. By
       default the time is midnight (first second of the day)
    """
    return datetime_to_datetime(_datetime.datetime.combine(date, time))


def get_datetime_now():
    """Return the current time in the UTC timezone. This creates an
       object that will be properly stored using datetime_to_string
       and string_to_datetime
    """
    return datetime_to_datetime(_datetime.datetime.now(
                                _datetime.timezone.utc))


def get_datetime_now_to_string():
    """Convenience function that returns the result of get_datetime_now
       as a string converted via datetime_to_string
    """
    return datetime_to_string(get_datetime_now())


def get_datetime_future(weeks=0, days=0, hours=0, minutes=0, seconds=0,
                        timedelta=None):
    """Return the datetime that is the supplied time in the future.
       This will raise an exception if the time is not in the future!
    """
    delta = _datetime.timedelta(weeks=weeks, days=days, hours=hours,
                                minutes=minutes, seconds=seconds)

    if timedelta is not None:
        if not isinstance(timedelta, _datetime.timedelta):
            raise TypeError("The delta must be a datetime.timedelta object")

        delta += timedelta

    if delta.total_seconds() < 5:
        raise ValueError(
            "The requested delta (%s) is not sufficiently far enough "
            "into the future!" % str(delta))

    return get_datetime_now() + delta


def string_to_datetime(s):
    """Return the datetime that had been encoded to the passed string
       via datetime_to_string. This string must have been created
       via 'datetime_to_string'
    """
    if isinstance(s, _datetime.datetime):
        return s
    else:
        d = _datetime.datetime.fromisoformat(s)
        return datetime_to_datetime(d)


def date_to_string(d):
    """Return the date that has been encoded to a string. This will
       write the date as a standard iso-formatted date. IF a datetime
       is passed then this will be in the
       UTC timezone (converting to UTC if the passed datetime
       is for another timezone)
    """
    if isinstance(d, _datetime.datetime):
        return datetime_to_datetime(d).date().isoformat()
    else:
        return d.isoformat()


def string_to_date(s):
    """Return a date from the string that has been encoded using
       'date_to_string'. This is only guaranteed to work for strings
       that were created using that function
    """
    d = _datetime.date.fromisoformat(s)
    return d


def time_to_string(t):
    """Return the time that has been encoded to a string. This will
       write the time as a standard iso-formatted time. If a datetime
       is passed then this will be in the
       UTC timezone (converting to UTC if the passed datetime
       is for another timezone)
    """
    if isinstance(t, _datetime.datetime):
        t = datetime_to_datetime(t)

        # guaranteed to be in the utc timezone, so write the
        # time without the unnecessary +00:00
        return t.replace(tzinfo=None).time().isoformat()
    else:
        if t.tzinfo is None:
            # assume UTC
            t = t.replace(tzinfo=_datetime.timezone.utc)
        elif t.tzinfo != _datetime.timezone.utc:
            from Acquire.ObjectStore import EncodingError
            raise EncodingError(
                "Cannot encode a time to a string as this time is "
                "not in the UTC timezone. Please convert to UTC "
                "before encoding this time to a string '%s'" % t.isoformat())

        # as the time is in UTC, we don't need the unnecessary +00:00
        return t.replace(tzinfo=None).isoformat()


def string_to_time(s):
    """Return a time from the string that was encoded by 'time_to_string'.
       This will only be guaranteed to produce valid output for strings
       produced using that function
    """
    t = _datetime.time.fromisoformat(s)

    if t.tzinfo is None:
        # assume this is a UTC time
        t = t.replace(tzinfo=_datetime.timezone.utc)
    else:
        t = t.astimezone(_datetime.timezone.utc)

    return t


def string_to_safestring(s):
    """Return a safe (base64) encoded version of 's'. This string
       has no special characters or spaces, thereby making it safe
       for use, e.g. as a filename or to save in a database
    """
    return _base64.b64encode(s.encode("utf-8")).decode("utf-8")


def safestring_to_string(s):
    """Return the original string encoded by string_to_safestring"""
    return _base64.b64decode(s.encode("utf-8")).decode("utf-8")


def list_to_string(l):
    """Return the passed list of items converted to a json string.
       All items should have the same type
    """
    j = []
    for item in l:
        j.append(item.to_data())

    return _json.dumps(j)


def string_to_list(s, C):
    """Convert the string encoded using list_to_string back to a list
        of objects of type C. Note that all objects must have the
        same type
    """
    items = []

    for val in _json.loads(s):
        items.append(C.from_data(val))

    return items


def dict_to_string(d):
    """Return the passed dict of items converted to a json string.
       All items should have the same type
    """
    j = {}
    for key, value in d.items():
        if value is None:
            j[key] = None
        else:
            j[key] = value.to_data()

    return _json.dumps(j)


def string_to_dict(s, C):
    """Convert the string encoded using dict_to_string back to a dict
        of objects of type C. Note that all objects must have the
        same type
    """
    items = {}

    for key, value in _json.loads(s).items():
        if value is None:
            items[key] = None
        else:
            items[key] = C.from_data(value)

    return items


def string_to_filepath(path):
    """This function cleans the passed path so that doesn't contain
       redundant slashes or '..' etc., so that all backslashes are forwards
       slashes, and that the trailing slash is removed
    """
    if path is None:
        return ""

    path = _os.path.normpath(path)

    # remove all ".." and "." from this path
    if path.find(".") != -1:
        parts = path.split("/")
        for i, part in enumerate(parts):
            if part == ".":
                parts[i] = None
            elif part == "..":
                if i == 0:
                    raise ValueError("You cannot start a path with '..'")
                part[i-1] = None
                part[i] = None

        path = _os.path.normpath("/".join(parts))

    return path


def string_to_filepath_parts(path):
    """Break the passed path into a list of the individual parts,
       e.g. /home/user/test/../something/./new.txt will return

       ['home', 'user', 'something', 'new.txt']
    """
    from os.path import split as _split

    path = string_to_filepath(path)

    dirs = []

    (root, part) = _split(path)

    dirs.append(part)

    while len(root) > 0:
        (root, part) = _split(root)

        if len(part) > 0:
            dirs.insert(0, part)
        else:
            break

    return dirs




import os as _os

if _os.getenv("PROFILE") == "1":
    profiling_code = True

    def start_profile():
        import cProfile as _cProfile
        pr = _cProfile.Profile()
        pr.enable()
        return pr

    def end_profile(pr, results):
        pr.disable()
        import tempfile as _tempfile
        from Acquire.ObjectStore import bytes_to_string as _bytes_to_string
        t = _tempfile.mktemp()
        pr.dump_stats(t)
        with open(t, "rb") as FILE:
            data = FILE.read()
        _os.unlink(t)
        results["profile_data"] = _bytes_to_string(data)

else:
    profiling_code = False

    def start_profile():
        return None

    def end_profile(profiler, results):
        return results

__all__ = ["start_profile", "end_profile"]
